if Array.length Sys.argv < 3 then (
  print_string ("
This is Blast's test driver generator.\n
Usage: " ^ Sys.argv.(0) ^ " <function name> <C files>.\n
<function name> is the name of the function under test.
<C files> is a list of C files containing the code to test.\n
http://www.eecs.berkeley.edu/~blast\n
");
  exit(1)
);

print_string "Parsing sources....\n";
flush stdout;
let func = Sys.argv.(1) in
let fnames =
  if Array.length Sys.argv >= 3 then
    Array.to_list (Array.sub Sys.argv 2 (Array.length Sys.argv - 2))
  else
    (let dir = Unix.opendir "." in
    let rec getFilenames () =
      try
	let fname = Unix.readdir dir in
	if (String.length fname >= 2 && compare ".c" (String.sub fname (String.length fname - 2) 2) == 0)
	    && compare "merged.c" fname != 0
	    && compare "instrumented.c" fname != 0 then
	  fname :: getFilenames ()
	else
	  getFilenames ()
      with End_of_file -> [] in
    let fnames = getFilenames () in
    Unix.closedir dir;
    fnames) in
let files = List.map Common.parseFile fnames in
let file = Mergecil.merge files "merged.c" in

let newName = match fnames with
  [fname] ->
    if String.length fname >= 2
	&& String.get fname (String.length fname - 2) = '.' then
      String.sub fname 0 (String.length fname - 2) ^ "_test.c"
    else
      failwith ("Unknown file type: " ^ fname)
| _ -> "merged_test.cpp" in

let file, call, numpar = Instrument.instrument func file in

(*print_string "\Before rmtmps:\n";
Common.printFile file;*)

Rmtmps.removeUnusedTemps file;

(*print_string "\nInstrumented code:\n";
Common.printFile file;*)

let outf = open_out newName in
  Printf.fprintf outf "// Begin: Added for test.\n";
  Printf.fprintf outf "#define TEST_CALL %s\n" call;
  Printf.fprintf outf "#define NUM_PAR %d" numpar;
  Printf.fprintf outf "
#define LINE_LENGTH 100000
#define DELIMITER \" \"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

char* gTestLocation;
int gTestLocReached; // Boolean.

int _blastReadTest() {
  char* lToken = strtok(NULL, DELIMITER);
  if (lToken == NULL) {  // No more test data.
    if (!gTestLocReached) { //  Test location reached first time.
      printf(\"Blast test driver: Location '%%s' reached.\\n\", gTestLocation);
      gTestLocReached = 1;
    }
    return 0;
  }
  return atoi(lToken);
}
// End: Added for test.
";

  Common.outputFile outf file;
  Printf.fprintf outf "
// Begin: Added for test.
int main(int argc, char *argv []) {
  if (argc < 2 || argc > 2)  
  {
    printf(\"%%s%%s%%s%%s%%s%%s%%s\",
      \"\\nThis is a test driver, generated by Blast's test driver generator.\\n\\n\",
      \"Usage: \", argv[0], \" <test file>.\\n\\n\",
      \"<test file> should contain one test vector per line.\\n\",
      \"The first word of a line is interpreted as the name of the test location.\\n\\n\",
      \"http://www.eecs.berkeley.edu/~blast\\n\\n\");
    exit(EXIT_FAILURE);
  }

  FILE* lTestFile = fopen (argv[1], \"r\");
  if (lTestFile == NULL) {
    printf(\"Blast test driver: Can't open input test file '%%s': %%s.\\n\", argv[1], strerror(errno));
    exit(EXIT_FAILURE);
  }

  char* lLine = (char*) calloc(LINE_LENGTH, sizeof(char));
  int* par = (int*) calloc(NUM_PAR, sizeof(int));

  while (fgets(lLine, LINE_LENGTH, lTestFile) != NULL) {
    printf(\"Blast test driver: Testing %%s\", lLine);
    if (strlen(lLine) == LINE_LENGTH-1) {
      printf(\"Blast test driver: Maximal test vecor size reached, increase 'LINE_LENGTH'.\\n\");
    }

    gTestLocation = strtok(lLine, DELIMITER);
    gTestLocReached = 0;
	
    // Handle parameters of function under test.
    int i;
    for(i=0; i<NUM_PAR; ++i) {
      par[i] = atoi(strtok(NULL, DELIMITER));
    }

    // Call function under test.
    TEST_CALL;

    _blastReadTest();  // Output of Location ... reached, if not yet done.
    printf(\"Blast test driver: Test done.\\n\");
  }
  free(par);
  free(lLine);
  fclose(lTestFile);
  printf(\"All tests successfully done.\\n\");
}
// End: Added for test.
";

  close_out outf;

  print_string "Done.\n"
