
(* need to link:
   misc.ml search.ml biDirectionalLabeledTree.ml *)

open Tester


(* Types similar to those used by the model checker *)
type node_data = { label : string; childno: int; level: int }
type operation = Empty | Atomic | Union
type tree_node = (node_data, operation) BiDirectionalLabeledTree.node
type iterator = tree_node Search.iterator


(* Add treewidth children to the parent node, if its level is less than treedepth - 1. *)
let add_children = fun (parent: tree_node) (treewidth: int) (treedepth: int)
                       iter ->
  let parent_data = (BiDirectionalLabeledTree.get_node_label parent)
  in let rec add_child = fun (childnumber : int) ->
	    let child_label = parent_data.label ^ "." ^ (string_of_int childnumber)
            in let node = BiDirectionalLabeledTree.create_child {label = child_label; childno = childnumber;
					     level = (parent_data.level+1)} Empty parent
	    in 
	      iter#add_element node;
	      if childnumber < (treewidth-1) then
		add_child (childnumber+1)
  in if parent_data.level<(treedepth-1) then
      add_child 0

(* Return a list of labels that correspond to the sequence generated by a dfs iterator *)
let dfs_sequence = fun (depth: int) (width: int) (logfn : string -> string -> unit) ->
  let unprocessed_nodes : iterator =  new Search.dfsiterator
  and tree_root = BiDirectionalLabeledTree.create_root {label = "0"; childno = 0; level = 0}
  and exp_sequence = ref []
   in unprocessed_nodes#add_element tree_root;
     logfn "expected" ("Sequence for tree of depth " ^ (string_of_int depth) ^ 
		     " width " ^ (string_of_int width) ^ ":");
     while unprocessed_nodes#has_next () do
       let n = unprocessed_nodes#next () in
         exp_sequence := (BiDirectionalLabeledTree.get_node_label n).label :: !exp_sequence;
	 logfn "expected" (BiDirectionalLabeledTree.get_node_label n).label;
	 add_children n width depth unprocessed_nodes
     done; List.rev !exp_sequence

(* Test the bounded dfs iterator *)
let bounded_iter_test = fun (tree_size: int) (bound: int) (logfn: string -> string -> unit) ->
  let unprocessed_nodes =  new Search.bounded_dfs_iterator bound
			     (fun n -> (BiDirectionalLabeledTree.get_node_label n).level+1)
  and tree_root = BiDirectionalLabeledTree.create_root {label = "0"; childno = 0; level = 0}
  and expected_sequence = (dfs_sequence (if (bound<tree_size) then bound else tree_size)
			                tree_size logfn)
  in let next_expected = ref expected_sequence
   in 
    unprocessed_nodes#add_element tree_root;
    logfn "actual"
      ("Sequence for tree of size " ^ (string_of_int tree_size) ^ " with bound " ^
       (string_of_int bound) ^ ":");
    while unprocessed_nodes#has_next () do
      let n = unprocessed_nodes#next () in
      let exp_label : string = List.hd !next_expected
      and actual_label : string = (BiDirectionalLabeledTree.get_node_label n).label in
	next_expected := List.tl !next_expected;
	logfn "actual" actual_label;
	assert_equal_string exp_label actual_label
	  "dequeued label did not match next in expected sequence";
	add_children n tree_size tree_size unprocessed_nodes
    done;
    assert_true (!next_expected = [])
      "more items in dfs sequence than bounded dfs sequence";
    if (bound < tree_size) then
       assert_true (unprocessed_nodes#was_bound_exceeded ())
	 "was_bound_exceed = false, should be true"
     else
       assert_false (unprocessed_nodes#was_bound_exceeded ())
	 "was_bound_exceed = true, should be false"

let suite =
  (new test_suite "bounded dfs tests" "tests for bounded dfs search (-bnddfs)"
     [ (new generic_test "bnddfs_2_2" "bnddfs option: equal bounds"
	  (fun ctx -> bounded_iter_test 2 2 ctx.logger#log_debug));
       (new generic_test "bnddfs_3_2" "bnddfs option: tree 1 deeper than bound"
	  (fun ctx -> bounded_iter_test 3 2 ctx.logger#log_debug));
       (new generic_test "bnddfs_4_2" "bnddfs option: tree 2 deeper than bound"
	  (fun ctx -> bounded_iter_test 4 2 ctx.logger#log_debug));
       (new generic_test "bnddfs_4_5" "bnddfs option: bound one deeper than tree"
	  (fun ctx -> bounded_iter_test 4 5 ctx.logger#log_debug));
       (new generic_test "bnddfs_1_1" "bnddfs option: single node"
	  (fun ctx -> bounded_iter_test 1 1 ctx.logger#log_debug));
     ])

let _ = test_script_main suite LogResults

