(*
 * Description:
 *
 * Copyright (c) 2000 by
 *  David Blei			blei@cs.berkeley.edu
 *  Chris Harrelson		chrishtr@cs.berkeley.edu
 *  Ranjit Jhala		jhala@cs.berkeley.edu
 *  Rupak Majumdar		rupak@cs.berkeley.edu
 *  George C. Necula	necula@cs.berkeley.edu
 *  Shree P. Rahul		sprahul@cs.berkeley.edu
 * 	Wes Weimer			weimer@cs.berkeley.edu
 *  Dror Weitz			dror@cs.berkeley.edu
 *   
 * All rights reserved.  Permission to use, copy, modify and distribute
 * this software for research purposes only is hereby granted, 
 * provided that the following conditions are met: 
 * 1.  Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 * 3. The name of the authors may not be used to endorse or promote products derived from 
 * this software without specific prior written permission. 
 *
 * DISCLAIMER:
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR IMPLIED 
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF 
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *)
module P = Proof
open VampyreErrormsg
open Proof
open Dag
open Satproc
open Num

(* State variables *)
let debug = ref false  (* debug flag *)
let supInfId = ref ""  (* what are we called? This is set in the initialize routine *)

type constant = Num.num
type vartype =
    RealNode of Dag.node
  | DebugNode of string

type table = Num.num array

type element =
    Constant of constant
  | TimesVar of constant * vartype
  | TimesMin of constant * (table option * expr) list
  | TimesMax of constant * (table option * expr) list
  | Infinity
  | NegInfinity
and expr = element list (* implicitly a sum of 1 or more things *)

exception SupInfException of (table option) * string

module Expr_set_type =
  struct
    type t = vartype
    let compare = compare
  end

module Expr_set = Set.Make(Expr_set_type)

let zero = Int 0
let one = Int 1
let minusOne  = Int (-1)


(* addTable -- a very imperative looking function that adds the corresponding
   entries of two tables
*)
let addTables (t1, t2) = 
	let a = Array.make (Array.length t1) (Int 0) in
	for i = 0 to (Array.length t1) - 1  do
	  a.(i) <- add_num t1.(i) t2.(i)
	done;
	a

let getTable sT = (* get the table from a table option *)
	match sT with
	Some t -> t
	| None -> raise (Failure "SupInf: getTable: No table!")

(******************************************************************************)
(* simp! *)
(******************************************************************************)

(* ***** *)
(* helper functions to print debugging messages: *)

let show_tables = true

(* convert a variable into a printable string *)
let string_of_var v =
  match v with
    DebugNode s -> s
  | RealNode d -> "{" ^ (string_of_int (d.Dag.id)) ^ "}"

let rec string_of_expr e =
  let rec string_of_table_help table i acc =
    if i = (Array.length table) then
      acc
    else
      if i > 0 then
	string_of_table_help table (i+1) (acc ^ "," ^ (string_of_num table.(i)))
      else
	string_of_table_help table (i+1) (string_of_num table.(0))
  and string_of_table table =
    if show_tables then
      match table with
	None -> "<e> "
      | Some tbl -> "<" ^ (string_of_table_help tbl 0 "") ^ "> "
    else
      ""
  and string_of_num' x =
    if not (eq_num x one) then
      string_of_num x
    else ""
  and helper c m is_min =
    let (_,min_str) =
      List.fold_left
	(function (flag,acc) ->
	  function (table, e') ->
	    if flag then
	      (true, acc ^ "," ^ (string_of_table table) ^ (string_of_expr e'))
	    else
	      (true, (string_of_table table) ^ (string_of_expr e') ^ acc))
	(false, "") m
    in
      if is_min then
	(string_of_num' c) ^ "min(" ^ min_str ^ ")"
      else
	(string_of_num' c) ^ "max(" ^ min_str ^ ")"
  in
  match e with
    first :: rest ->
      let first_string =
	match first with
	  Constant c -> string_of_num c
	| Infinity -> "inf"
	| NegInfinity -> "n_inf"
	| TimesVar (c, x) -> (string_of_num' c) ^ (string_of_var x)
	| TimesMin (c, m) -> helper c m true
	| TimesMax (c, m) -> helper c m false
      in
	(match rest with
	  _ :: _ -> first_string ^ " + " ^ (string_of_expr rest)
	| _ -> first_string ^ (string_of_expr rest))
  | [] -> ""

(* end helper functions for printing debugging *)
(* ***** *) 

let rec mult_by const el =
  let map_func c (table, l) =
    match table with
      None -> (None, List.map (mult_by c) l)
    | Some t ->
	let c' = if lt_num c zero then minus_num c else c
	and a = Array.copy t
	in
	  for i = 0 to (Array.length a) - 1 do
	    a.(i) <- mult_num a.(i) c'
	  done;
	  (Some a, List.map (mult_by c) l)
  in
    match el with
      TimesMax (c,l) ->
	let res = mult_num const c
	in
	  if lt_num res zero then
	    TimesMin (one, List.map (map_func res) l)
	  else
	    TimesMax (one, List.map (map_func res) l)
    | TimesMin (c,l) ->
	let res = mult_num const c
	in
	  if lt_num res zero then
	    TimesMax (one, List.map (map_func res) l)
	  else
	    TimesMin (one, List.map (map_func res) l)
    |	TimesVar (c, v) ->
	let res = mult_num const c
	in TimesVar (res, v)
    |	Constant c -> Constant (mult_num const c)
    |	Infinity -> if lt_num const zero then NegInfinity else Infinity
    |	NegInfinity -> if lt_num const zero then Infinity else NegInfinity

(* distributes + over min and max, and converts min to max and max to min if necessary because of negative coefficients. *)
(*  assumes you have already run const_collect_min_convert *)
let rec distribminmax e =
  let (first_pass, first_pass_rest, first_pass_flag) = (* push multiplicative constant inside *)
    match e with
      ((TimesMax _) as first)::rest ->
	(Some first, rest, true)
    | (TimesMin _ as first)::rest ->
	(Some first, rest, true)
    | _ -> (None, e, false) 
  in
  (* add the rest of the terms *)
    if first_pass_flag then
	  match first_pass with
	    Some (TimesMax (c, l)) ->
	      let added_val = List.map (function (table, sum) -> (table, List.append sum first_pass_rest)) l
	      in
		[TimesMax (c, List.map (function (table, e) -> (table, distribminmax e)) added_val)]
	  | Some (TimesMin (c, l)) ->
	      let added_val = List.map (function (table, sum) -> (table, List.append sum first_pass_rest)) l
	      in
		[TimesMin (c, List.map (function (table, e) -> (table, distribminmax e)) added_val)]
	  | _ -> raise (Failure "supInf: distribminmax: bug in distribminmax")
    else
      first_pass_rest

(* pushes constants inside min and max terms *)
let rec mult_minmaxes_by_const e =
  match e with
    (TimesMin (c, m))::rest -> (mult_by c (TimesMin (one, m)))::(mult_minmaxes_by_const rest)
  | (TimesMax (c, m))::rest -> (mult_by c (TimesMax (one, m)))::(mult_minmaxes_by_const rest)
  | first::rest -> first::(mult_minmaxes_by_const rest)
  | _ -> []

let add_arrays a1 a2 =
  match (a1, a2) with
    (None, None) -> None
  | (Some _, None) -> a1
  | (None, Some _) -> a2
  | (Some t1, Some t2) ->
      let a = Array.make (Array.length t1) (Int 0)
      in (* hideous!  terrible! *)
	(for i = 0 to (Array.length t1) - 1 do
	  a.(i) <- add_num t1.(i) t2.(i)
	done;
    Some a)

(* flattens with respect to min and max.  assumes you have already run distribminmax, so that every sum has any mins or maxes only *)
(* at the beginning (i.e. is of the form [TimeMax (c, m); ...] or [TimesMin (c, m); ...] where ... contains no min or max).  It should *)
(* also have the constants c in TimesMin(c, m) etc. already pushed inside. *)
let rec flatten e =
  let add_table_list table m' =
    List.map
      (function
	  ((Some tbl) as t, term) -> (add_arrays table t, term)
	| (None, term) -> (table, term)) m'
  in
    match e with
      [TimesMax (c, m)] ->
	[TimesMax (c,
		   List.fold_left
		     (function acc ->
		       function (table, term) ->
			 match flatten term with
			   [TimesMax (c, m')] ->
			     (add_table_list table m') @ acc
			 | [TimesMin _] -> raise (Failure "sup-inf: bug in flatten")
			 | term' -> (table, term') :: acc)
		     [] m)]
    | [TimesMin (c, m)] ->
	[TimesMin (c,
		   List.fold_left
		     (function acc ->
		       function (table, term) ->
			 match flatten term with
			   [TimesMin (c, m')] ->
			     (add_table_list table m') @ acc
			 | [TimesMax _] -> raise (Failure "sup-inf: bug in flatten #2")
			 | term' -> (table, term') :: acc)
		     [] m)]
    | _ -> e


(* This function turns an expression of the form min(..) + .. + min(..) + e (where the ".."'s and e do not contain min) into one of the form *)
(* min(..) + e.  Of course, min(..) could also be max(..), but you can't mix the two. *)
let rec make_multiple_minmaxes_into_one e =
  let rec helper (min, nonmin) term =
    match term with
      (TimesMax _) | (TimesMin _) ->
	(term::min, nonmin)
    | _ -> (min, term::nonmin)

  and mk_together (st1, e1) (st2, e2) =
    (add_arrays st1 st2, (e1 @ e2))

  and combine_two_mins min1 min2 =
    List.fold_left
      (function acc -> function el1 ->
	acc @ (List.map
		 (function el2 ->
		   mk_together el1 el2) min2))
      [] min1

  and combine_mins mins =
    match mins with
      (TimesMin (c, a))::(TimesMin (_, b))::rest -> combine_mins ((TimesMin (c, combine_two_mins a b)) :: rest)
    | (TimesMax (c, a))::(TimesMax (_, b))::rest -> combine_mins ((TimesMax (c, combine_two_mins a b)) :: rest)
    | _ -> mins

  and recurse e =
    match e with
      (TimesMin (c,m))::rest ->
	(TimesMin (c, List.map (function (t,e) -> (t,make_multiple_minmaxes_into_one e)) m))::rest
    | (TimesMax (c,m))::rest ->
	(TimesMax (c, List.map (function (t,e) -> (t,make_multiple_minmaxes_into_one e)) m))::rest
    | first::rest -> raise (Failure "bug in make_multiple_minmaxes_into_one")
    | _ -> e
  in
    let (min_terms, non_min_terms) = List.fold_left helper ([], []) e
    in
      (recurse (combine_mins min_terms)) @ non_min_terms

(* 1. collects all constants in a sum to one constant *)
(* 2. moves any min or max to the front of the sum list *)
(* 3. removes terms with a zero coefficient *)
let rec const_collect_min_convert e =
  let try_to_eval term =
    let rec try_helper flag cur_min cur_expr e =
      match e with
	(table, [Constant a])::rest ->
	  (match cur_min with
	    Some (_, [Constant num]) ->
	      if (flag && (lt_num a num)) || ((not flag) && (gt_num a num)) then (* if I'm better than the current best, then change to me *)
		try_helper flag (Some (table, [Constant a])) cur_expr rest
	      else try_helper flag cur_min cur_expr rest
	  | None -> try_helper flag (Some (table, [Constant a])) cur_expr rest
	  | Some (_,[Infinity]) ->
	      if flag then (* min *)
		try_helper flag (Some (table, [Constant a])) cur_expr rest
	      else (* max *)
		raise (Failure "sup-inf: bug in const_collect_min_convert #4")
	  | Some (_,[NegInfinity]) ->
	      if flag then (* min *)
		raise (Failure "sup-inf: bug in const_collect_min_convert #5")
	      else (* max *)
		try_helper flag (Some (table, [Constant a])) cur_expr rest
	  | _ -> raise (Failure "sup-inf: bug in const_collect_min_convert #1"))
      | (table, [Infinity])::rest ->
	  if flag then (* min *)
	    (match cur_min with
	      Some _ -> try_helper flag cur_min cur_expr rest
	    | None -> try_helper flag (Some (table, [Infinity])) cur_expr rest)
	  else (* max *)
	    (Some (table, [Infinity]), [])
      | (table, [NegInfinity])::rest ->
	  if flag then (* min *)
	    (Some (table, [NegInfinity]), [])
	  else (* max *)
	    (match cur_min with
	      Some _ -> try_helper flag cur_min cur_expr rest
	    | None -> try_helper flag (Some (table, [NegInfinity])) cur_expr rest)
      | first::rest -> try_helper flag cur_min (first::cur_expr) rest
      | [] -> (cur_min, cur_expr)
    in
      match term with
	TimesMax (c, l) ->
	  (match try_helper false None [] l with
	    (Some const_term, expr) -> TimesMax (c, const_term::expr)
	  | (None,_) -> term)
      | TimesMin (c, l) ->
	  (match try_helper true None [] l with
	    (Some const_term, expr) -> TimesMin (c, const_term::expr)
	  | (None,_) -> term)
      |	_ -> term
  in
    let const_sum =
      List.fold_left
	(function (acc,flag) ->
	  function
	      (Constant c) ->
		(match acc with
		  Constant const ->
		    (Constant (add_num const c), true)
		| (Infinity | NegInfinity) ->
		    (acc, flag)
		| _ -> (acc, flag))
	    | Infinity ->
		(match acc with
		  NegInfinity ->
		    (print_string "hello\n";
		     raise (Failure "sup-inf: bug in const_collect_min_convert #2"))
		| _ -> (Infinity, true))
	    | NegInfinity ->
		(match acc with
		  Infinity -> raise (Failure "sup-inf: bug in const_collect_min_convert #3")
		| _ -> (NegInfinity, true))
	    | _ -> (acc,flag))
	(Constant zero,false) e
    and non_const_entries =
      List.fold_left
	(function (l, m) ->
	  function
	      ((Constant _) | Infinity | NegInfinity) -> (l, m)
	    | TimesMin (c,m') ->
		let retval = try_to_eval (TimesMin (c,List.map
						      (function (t,e') -> (t, const_collect_min_convert e'))
						      m'))
		in
		  (match retval with
		  | TimesMin _ -> (l, Some retval)
		  | _ -> (retval::l, m))
	    | TimesMax (c,m') ->
		let retval = try_to_eval (TimesMax (c,List.map
						      (function (t,e') -> (t, const_collect_min_convert e'))
						      m'))
		in
		  (match retval with
		    TimesMax _ -> (l, Some retval)
		  | _ -> (retval::l, m))
	    | a -> (a::l, m))
	([],None) e
    in
      match const_sum with
	(_,false) ->
	  (match non_const_entries with
	    (l, Some m) -> m::(List.rev l)
	  | (l, None) -> List.rev l)
      | (acc,true) ->
	  (match non_const_entries with
	    (l, Some m) -> m::acc::(List.rev l)
	  | (l, None) -> acc::(List.rev l))

let rec merge a1 a2 = (* merge two lists, only keeping one copy of each element *)
   match (a1,a2) with
     ([], a) -> a
   | (a, []) -> a
   | (e::l, b) -> if (List.mem e b) then
                    merge l b
                  else
                    e::(merge l b) 

(* Return a list of variables that occur in the expr c *)
let rec getVars c = 
  match c with
    [] -> [] 
  | e::rest -> 
      let vl = (getVarsFromElement e ) in
        (merge vl (getVars rest))
and 
    getVarsFromElement e = 
  match e with
    Constant c -> []
  | Infinity -> []
  | NegInfinity -> []
  | TimesVar(_, v) -> [v]
  | TimesMin(_, ex) ->  List.fold_left merge [] (List.map (function (t,e') -> getVars e') ex)
  | TimesMax(_, ex) ->  List.fold_left merge [] (List.map (function (t,e') -> getVars e') ex)

let rec mergeVars var exp i =
    match exp with
      [] -> if(i<>/ (Int (0))) then (TimesVar(i, var))::[] else []
    | a::rest -> (
        match a with
          TimesVar(j, v) ->
	    if (v=var) then (mergeVars var rest (i +/ j))
            else TimesVar(j,v)::(mergeVars var rest i)
        | TimesMax(j, e) -> TimesMax(j, List.map (function (t,e') -> (t,mergeAllVars e')) e)::(mergeVars var rest i)
        | TimesMin(j, e) -> TimesMin(j, List.map (function (t,e') -> (t,mergeAllVars e')) e)::(mergeVars var rest i)
        | r -> r::(mergeVars var rest i)
       )
and mergeAllVars p =
  let f e v = (mergeVars v e (Int 0))
  in
    let varl = (getVars p)
    in
      List.fold_left f p varl

(* replaces any silly empty lists that have popped up due to mergeAllVars with zero  i.e. max(1,[],0) -> max(1,0,0). *)
let remove_empty_minmax_els e =
  let helper m =
    List.map
      (function (table, []) -> (table, [Constant zero])
	| a -> a) m
  in
    match e with
      [TimesMax (c, m)] -> [TimesMax (c, helper m)]
    | [TimesMin (c, m)] -> [TimesMin (c, helper m)]
    | _ -> e
   
let simp e =
  let v = 
  const_collect_min_convert
    (remove_empty_minmax_els
       (mergeAllVars 
	  (flatten
	     (distribminmax
		(const_collect_min_convert
		   (make_multiple_minmaxes_into_one
		      (mult_minmaxes_by_const e)))))))
  in
  match v with
  [] -> [Constant (Int 0)]
  | _ -> v

(* end simp *)


(* Other simple operations on expressions *)

(* negate: negate an expression *)
(* [Check] *)
let rec negate e =
  match e with
    [] -> []
    | el::rest -> (
        match el with
          Constant c -> Constant (mult_num minusOne c)
	| TimesVar (c,v) -> TimesVar((mult_num minusOne c), v)
	| TimesMax (c,v) -> TimesMax((mult_num minusOne c), v)
	| TimesMin (c,v) -> TimesMin((mult_num minusOne c), v)
	| Infinity -> NegInfinity
	| NegInfinity -> Infinity
    ) :: (negate rest)


let rec contains var expr =
  match expr with
    a::rest ->
      (match a with
	TimesVar (c,v) ->
	  if v = var then true
	  else contains var rest
      |	TimesMin (c,l) ->
	  if List.fold_left (function acc -> function (t, term) -> acc || (contains var term)) false l then true
	  else contains var rest
      |	TimesMax (c,l) ->
	  if List.fold_left (function acc -> function (t, term) -> acc || (contains var term)) false l then true
	  else contains var rest
      |	Infinity -> contains var rest
      |	NegInfinity -> contains var rest
      |	Constant _ -> contains var rest)
  | [] -> false      

(* solve_for: Solve a set of constraints for a variable. *)
(* solve_for_upper returns expressions exp of the form x>=exp *)
(* solve_for_lower returns expressions exp of the form x<=exp *)
(* s: set of variables *)
(* var: variable to solve for *)
(* return a list of expressions *)

(* uses the following help functions *)
let rec is_pos_is_neg flag var linexp = (* linexp is a linear expression, flag indicates positive or negative *)
	(* assume each variable occurs exactly once *)
  match linexp with
    a::rest ->
      (match a with
	TimesVar (c, v) ->
	  if var = v then
	    if flag then
	      ((gt_num c zero) || (is_pos_is_neg flag var rest))
	    else
	      ((lt_num c zero) || (is_pos_is_neg flag var rest))
	  else
	    is_pos_is_neg flag var rest
      | _ -> is_pos_is_neg flag var rest)
  | [] -> false

let is_pos = is_pos_is_neg true
let is_neg = is_pos_is_neg false

let rec solve_for_upper_or_lower flag var n i =
  function exp::rest ->
    let var_term =
      List.fold_left
	(function acc ->
	  function
	      (TimesVar (c,v)) as term ->
		if v = var then
		  if (flag && (lt_num c zero)) || ((not flag) && (gt_num c zero)) then
		    (Some term)
		  else acc
		else
		  acc
	    | _ -> acc)
	None exp
    and non_var_terms =
      List.fold_left
	(function acc ->
	  function ((TimesVar (c,v)) as term) ->
	    if v = var then acc else term::acc
	    | term -> term::acc)
	[] exp
    in
      (match var_term with
	None -> solve_for_upper_or_lower flag var n (i+1) rest
      |	Some (TimesVar (c,v)) ->
	  (match non_var_terms with
	    [] ->
	      if (flag && (gt_num c zero)) || ((not flag) && (lt_num c zero)) then
		solve_for_upper_or_lower flag var n (i+1) rest
	      else
		let a = Array.make n zero in
		  (if lt_num c zero then
		    a.(i) <- (div_num one (minus_num c))
		  else
		    a.(i) <- (div_num one c);
		   (Some a, [Constant zero]) :: (solve_for_upper_or_lower flag var n (i+1) rest))
	  | terms ->
	      let a = Array.make n zero in
		(if lt_num c zero then
		  a.(i) <- (div_num one (minus_num c))
		else
		  a.(i) <- (div_num one c);
		 (Some a, List.map
		    (function
			(TimesVar (c',v)) ->
			  if flag then
			    TimesVar (div_num c' (minus_num c), v)
			  else
			    TimesVar (div_num (minus_num c') c, v)
		      | (Constant c') ->
			  if flag then
			    (Constant (div_num c' (minus_num c)))
			  else
			    (Constant (div_num (minus_num c') c))
		      | _ -> raise (Failure "sup-inf: bug in solve_for_upper_or_lower"))
		    terms) :: (solve_for_upper_or_lower flag var n (i+1) rest)))
      |	_ -> raise (Failure "sup-inf: bug in solve_for_upper_or_lower"))
    | _ -> []

let solve_for_upper v s = solve_for_upper_or_lower true v (Array.length s) 0 (Array.to_list s)

let solve_for_lower v s = solve_for_upper_or_lower false v (Array.length s) 0 (Array.to_list s)

(* get_x_terms returns the term containing the variable x and the terms not containing it.  it assumes there is only term *)
(* containing x *)
let get_x_terms x e = (* e is assumed to be a linear form (with no mins and maxes) *)
  let x_terms =
    List.fold_left
      (function acc ->
	function
	    (Constant c) -> acc
	  | (TimesVar (c,var)) as term ->
	      if var = x then Some term
	      else acc
	  | _ -> acc)
      None e
  and non_x_terms =
    List.fold_right
      (function term ->
	function acc ->
	  match term with
	    Constant c -> term::acc
	  | TimesVar (c,var) ->
	      if var = x then acc else term::acc
	  | _ -> term::acc)
      e []
  in
    (x_terms, non_x_terms)

let eq_var var_str exp =
  match exp with
    [TimesVar (c,v)] -> (v = var_str) && (eq_num c one)
  | _ -> false

let upper s var = 
  let explist = (solve_for_upper var s) in
    let lenexplist = List.length(explist) in
      if (lenexplist >= 1) then
	[TimesMin (one, explist)]
      else [TimesMin (one, [(None, [Infinity])])] (* what am I supposed to do here? *)

let lower s var = 
  let explist = (solve_for_lower var s) in
    let lenexplist = List.length(explist) in
      if (lenexplist >= 1) then
	[TimesMax (one, explist)]
      else [TimesMax (one, [(None, [NegInfinity])])] (* what am I supposed to do here? *)

let rec extract_var sum cur_sum = (* takes out the first variable term it finds and returns everything else as a separate sum *)
  match sum with
    ((TimesVar _) as first)::rest -> (Some first, (List.rev cur_sum) @ rest)
  | first::rest -> extract_var rest (first::cur_sum)
  | [] -> (None, cur_sum)

let rec string_space l =
  if l >= 4 then
    "     " ^ (string_space (l-4))
  else match l with 3 -> "   " | 2 -> "  " | 1 -> " " | _ -> ""

let rec sup_or_inf is_sup s indent h j =
  let
      (sup_inf, inf_sup, supp_inff, upper_lower) =
    if is_sup then
      (if !debug then
	(print_string (string_space indent);
	 print_string ( "sup( " ^ (string_of_expr j) ^ " )\n"));
       (sup_or_inf true s (indent + 2), sup_or_inf false s (indent + 2), supp, upper s))
    else
      (if !debug then
	(print_string (string_space indent);
	 print_string ("inf( " ^ (string_of_expr j) ^ " )\n"));
       (sup_or_inf false s (indent + 2), sup_or_inf true s (indent + 2), inff, lower s))
  in
    let result =
      match j with
	[] -> raise (Failure "supInf: sup_or_inf: bug in sup_inf")
      | [Constant c] -> j
      | [TimesVar (c,v)] ->
	  if eq_num c one then
	    (if (Expr_set.mem v h) then
	      j
	    else
	      let q = upper_lower v
	      in
		let z = sup_inf (Expr_set.add v h) q
		in
		  supp_inff v (simp z) None)
	  else if lt_num c zero then
	    List.map (mult_by c) (inf_sup h [TimesVar (one, v)])
	  else
	    List.map (mult_by c) (sup_inf h [TimesVar (one, v)])
      | [TimesMin (c,m)] ->
	  if (lt_num c zero) then
	    List.map (mult_by c) (inf_sup h [TimesMin (one, m)])
	  else
	    [TimesMin (c, List.map (function (table, e) -> (table, sup_inf h e)) m)]
      | [TimesMax (c,m)] ->
	  if (lt_num c zero) then
	    List.map (mult_by c) (inf_sup h [TimesMax (one, m)])
	  else
	    [TimesMax (c, List.map (function (table, e) -> (table, sup_inf h e)) m)]
      | [Infinity] ->
	  j
      | [NegInfinity] ->
	  j
      | otherwise ->
	  let (first_var_mult, rest) = extract_var otherwise []
	  in
	    (match first_var_mult with
	      Some (TimesVar (c,var)) ->
		let b' = sup_inf (Expr_set.add var h) rest
		in
		  if (contains var b') then
		    let j' = simp ((TimesVar (c,var))::b')
		    in
		      sup_inf h j'
		  else
		    (sup_inf h [TimesVar (c,var)]) @ b'
	    | None ->
		otherwise (* what am I supposed to do here anyway?? *)
	    | _ -> raise (Failure "sup-inf: bug in extract_var"))
    in
      (if !debug then
	(print_string (string_space indent);
	 print_string ("-> " ^ (string_of_expr result) ^ "\n"));
       result)
and supp x y table =
  if (eq_var x y) then
    [Infinity]
  else
    (match y with
      [] -> raise (Failure "sup-inf: bug in supp")
    | [Constant b] -> y
    | [Infinity] -> y
    | [NegInfinity] -> y
    | [TimesMin (c,m)] -> [TimesMin (c, List.map (function (t,e) -> (t,supp x e t)) m)] (* after simp, c should always be a one *)
    | _ ->
	let (x_term, non_x_terms) = get_x_terms x y
	in
	  match x_term with
	    None ->
	      (match non_x_terms with
		[] -> raise (Failure "sup-inf: bug in supp #2")
	      |	_ -> non_x_terms)
	  | Some (TimesVar (b,var)) ->
	      if (gt_num b one) then [Infinity]
	      else
		(match non_x_terms with
		  [] ->
		    if (lt_num b one) then
		      [Constant zero]
		    else
		      [Infinity]
		| _ ->
		    if (lt_num b one) then
		      List.map (mult_by (div_num one (sub_num one b))) non_x_terms
		    else
		      match non_x_terms with
			[Constant num] ->
			  if (lt_num num zero) then (* found a contradictory equation immediately *)
			    raise (SupInfException (table, "supp wanted to return NegInfinity"))
			  else [Infinity]
		      | _ -> [Infinity])
	  | _ -> raise (Failure "sup-inf: unexpected situation in supp"))
and inff x y table =
  if (eq_var x y) then
    [NegInfinity]
  else
    (match y with
      [Constant b] -> y
    | [Infinity] -> y
    | [NegInfinity] -> y
    | [TimesMax (c,m)] -> [TimesMax (c, List.map (function (t,e) -> (t,inff x e t)) m)]
    | _ ->
	let (x_term, non_x_terms) = get_x_terms x y
	in
	  match x_term with
	    None ->
	      (match non_x_terms with
		[] -> raise (Failure "sup-inf: bug in inff")
	      |	_ -> non_x_terms)
	  | Some (TimesVar (b,x')) ->
	      if (gt_num b one) then [NegInfinity]
	      else
		(match non_x_terms with
		  [] ->
		    if (lt_num b one) then
		      [Constant zero]
		    else
		      [NegInfinity]
		| _ ->
		    if (lt_num b one) then
		      List.map (mult_by (div_num one (sub_num one b))) non_x_terms
		    else
		      match non_x_terms with
			[Constant num] ->
			  if (gt_num num zero) then
			    raise (SupInfException (table, "inff wanted to return Infinity"))
			  else [NegInfinity]
		      |	_ -> [NegInfinity])
	  | _ -> raise (Failure "sup-inf: unexpected situation in inff"))

let sup eqns var =
  let res = simp (sup_or_inf true (Array.map (function e -> simp e) eqns) 0 Expr_set.empty var)
  in
    (if !debug then print_string "------------------------\n";
     res)
let inf eqns var =
  let res = simp (sup_or_inf false (Array.map (function e -> simp e) eqns) 0 Expr_set.empty var)
  in
    (if !debug then print_string "------------------------\n";
     res)

(* [Rupak] Changes added *)
(** Integrate with the rest of the system **)

(* Generate unique identifier numbers for messages *)
let nextid = ref (-1)
let nextId () = 
    nextid := !nextid + 1;
	!nextid
let currentId () =
	!nextid
let resetId (i) = (nextid := i)

let getNumberOfConstraints () = (!nextid + 1)  (* remember we start from index 0 *)


(** Snapshot and Undo **)
(*
  The undoStack is a list that contains the linear constraints interspersed with
  Snapshot markers. The linear constraints are added in the assume function from
  external messages; the Snapshot markers are added by calls to snapshot().

  An undo() will then
  remove all constraints added since the last snapshot
*)
type undoElem = 
    Snapshot  of int  (* mark a snapshot *)
    | Constraint of expr * Satproc.message * bool * (unit -> Proof.proof) * int	
									(* expr is the expression, 
									   message is the original message,
									   (function () -> proof) lazily supplies the proof, 
									   boolean is needed to remember which part of the equality this constraint is (>= or <=),
									and the final int is the constraint number
                                    *)

(* undoStack implemented as a list *)
let undoStack : undoElem list ref = ref []


let snapshot () = (* simply put a Snapshot marker on the undoStack 
					 and remember the Id number
                  *)
    undoStack := (Snapshot (currentId()))  :: !undoStack
   
(* Discard all new constraints since last snapshot. Discard all sup and
   inf values computed since last snapshot.

   Raise an exception if an undo is attempted on an undoStack without a
   prior snapshot()
*)
let undo () =
    if (!debug) then ignore
	  (VampyrePretty.printf "SupInf : undo : removing constraints from undoStack\n");
    let rec removeAdditions = function
	    (Snapshot snap) :: rest -> (undoStack := rest; resetId (snap))
    |   _ :: rest -> removeAdditions rest
	|   [] -> VampyreErrormsg.s (VampyreErrormsg.bug "SupInf.undo : Popping empty undo stack! An undo() without a prior snapshot!")
	in
	    removeAdditions !undoStack


(** END snapshot() and undo() **)








(* Filter the undo list to only keep the constraints.
   The resulting list will be sent to sup and inf
*)
let rec filterUndoList e = 
    match e with
    [] -> []
	|  (Snapshot _) :: rest -> (filterUndoList rest)
	|  (Constraint (c,s,b,p,id)) :: rest -> c::(filterUndoList rest)

let rec filterUndoListForMessage e = 
    match e with
    [] -> []
	|  (Snapshot _) :: rest -> (filterUndoListForMessage rest)
	|  (Constraint (c,s,b,p,id)) :: rest -> (s,b)::(filterUndoListForMessage rest)

(* get the message associated with a constraint from its id   *)
(* ll will be instantiated at the top level with (!undoStack) *)
(* This will be used by the proof generation procedures to    *)
(* get the proofs for the messages used in generating a       *)
(* contradiction *)
let rec findMessage i ll =
	match (ll) with
      [] ->
	(if !debug then
	  ignore (VampyrePretty.printf "findMessage: Raising contradiction on %d\n" i));
        raise (Failure "SupInf: findMessage: Message corresponding to id Not found")
    | (Snapshot _)::rest -> findMessage i rest
    | (Constraint (a,b,bit,p,d))::rest -> if (d=i) then (Constraint (a,b,bit,p,d))
                                      else (findMessage i rest)

let findConstraint i il =
	let c = findMessage i il in
	match (c) with
	Constraint (a,_,_,_,_) -> a
	| _ -> raise (Failure "SupInf: findConstraint: findMessage returned strange value")
	
(*
   expOfDag

*)
let expOfDag node =
  match nodeVal node with
    DExp (e) -> e
  | DAtom _ -> failwith "expOfDag: tried to get expression of an atom node"

let getLeft c = 
	match (c) with
	| Snapshot _ -> raise (Failure "SupInf: getLeft: This should be called only with constraints")
	| Constraint (_,b,_,_,_) -> (match b.body with
			Literal (l) -> (match (Dag.nodeVal (List.nth l.args 0)) with
				Dag.DAtom _ -> raise (Failure "SupInf: getLeft : Bug")
			  | Dag.DExp e -> e)
		|   Equality (a,b) -> (match 	(Dag.nodeVal (a)) with
    					Dag.DAtom _ -> raise (Failure "SupInf: getLeft : Bug")
    				| 	Dag.DExp e -> e)

			(* Note that this was a valid equality *)
		)

let getRight c = 
	match (c) with
	| Snapshot _ -> raise (Failure "SupInf: getLeft: This should be called only with constraints")
	| Constraint (_,b,_,_,_) -> (match b.body with
			Literal (l) -> (match (Dag.nodeVal (List.nth l.args 1)) with
				Dag.DAtom _ -> raise (Failure "SupInf: getLeft : Bug")
			  | Dag.DExp e -> e)
		|   Equality (a,b) -> (match 	(Dag.nodeVal (b)) with
    					Dag.DAtom _ -> raise (Failure "SupInf: getLeft : Bug")
    				| 	Dag.DExp e -> e)

			(* Note that this was a valid equality *)
		)



(*** BEGIN  Proof generation *)

(* Assuming that sup and inf return tables mapping constraints to coefficients,
   generate proof for contradiction
   t = table of coefficients
*)

(* getPf: gets the proof of the i th message *)
let getPf i = 
	(if (!debug) then (ignore (VampyrePretty.printf("SupInf: In getPf -- getting proof\n"))));
	let cons = (findMessage i !undoStack) in 
	match cons with
	  Snapshot _ -> raise (Failure "SupInf: getPf: Unexpected value returned by findMessage")
    | Constraint (a,b,_,p,d) -> p()


(* makeProofList: Takes a table of coefficients, and for each element with
   nonzero coefficient creates a pair (pf linearConstraint, coeff), and
   returns a list of such pairs
*)
let makeProofList tlist =
    let rec makePfListWorker tlist i =
	match tlist with
      [] -> []
    | coeff::rest -> if (coeff =/ (Int 0)) then (makePfListWorker rest (i+1)) 
                     else begin
						match coeff with
						Int(c) -> (getPf i, c)::(makePfListWorker rest (i+1))						
						| Big_int(c) -> (getPf i, Big_int.int_of_big_int(c))::(makePfListWorker rest (i+1))
						| _ -> raise (Failure "SupInf: makePfListWorker: At this stage coefficients should be integral")
                     end
    in
    makePfListWorker tlist 0

(* return coefficient table (as a list) with integral coefficients *)
let makeIntegerCoeffs t = 
  let  leastCommonMult lst = (List.fold_left
    	(function x -> function y -> if (Big_int.eq_big_int y Big_int.zero_big_int) then x else
					(Big_int.div_big_int (Big_int.mult_big_int x y) (Big_int.gcd_big_int x y)))
    Big_int.unit_big_int lst)
  in
  let tl = Array.to_list t in
  let coef_lst = List.map (function e -> (Num.ratio_of_num e)) tl in (* Num.num list *)
  let lcm = leastCommonMult (List.map (function c -> Ratio.denominator_ratio c) coef_lst) in (* Big_int.big_int *)
  let int_coefs = List.map (function c -> (Ratio.big_int_of_ratio (Ratio.mult_big_int_ratio lcm c))) coef_lst in
  List.map (function c-> num_of_big_int c) int_coefs

(* Given the coefficients produced by sup-inf, this function adds up the constraints
   weighing them by the coefficients.
   The final answer should be a negative integer.

   Useful for debugging/regression testing
*)
let getEliminant coeffTable = 
  let ilist = makeIntegerCoeffs coeffTable in
  let rec getConstWorker ilist i =
	(match ilist with
		[] -> []
     |  a::rest -> if(a=/ Int(0)) then (getConstWorker rest (i+1)) else
					simp((List.map (mult_by (a)) (findConstraint i !undoStack))) @ (getConstWorker rest (i+1))
	)
  in
  let en = simp (getConstWorker ilist 0) in
    (if (!debug) then 
      (ignore (VampyrePretty.printf("SupInf: getEliminant: printing expression after addition:\n"))));
  if !debug then (ignore(print_string (string_of_expr en)));
  en
(* 
   makeCombProof: Constructs the sum of the inequalities using the coefficients
   in the table (the coefficients are first converted to integers)
*)
let makeCombProof t =
  (if (!debug) then
  (ignore (VampyrePretty.printf "SupInf: makeCombProof: Size of t = %d\n" (Array.length t))));
  let integer_t_list = makeIntegerCoeffs t in
  let genPf restOfPf (pf, coeff) =
	Proof.G ("geq*>=", [I coeff; P pf; P restOfPf]) in
  let pfList = (makeProofList integer_t_list) in
  if (!debug) then
	(ignore (List.iter (function (j,i) -> ignore (VampyrePretty.printf "Coefficient after makeProofList %d\n" (i))) pfList));
   List.fold_left genPf
    (Proof.G ("geq0",[])) pfList


let getExpFromMsg msg bit = 
	match (msg.body) with 
	  Literal l -> (
		let d1 = (List.nth l.args 0) 
		and d2 = (List.nth l.args 1) in
		let e1 = expOfDag d1
		and e2 = expOfDag d2
		in
		if (msg.goal) then
			match l.op with
			Lt  -> Input.Minus(e1, e2)
		|	Gt  -> Input.Minus(e2, e1)
		|	Geq -> Input.Minus(e2, Input.Plus(e1, Input.Int 1))
		|	Leq -> Input.Minus(e1, Input.Plus(e2, Input.Int 1))
		|	Eq	-> if bit then Input.Minus(e1, Input.Plus(e2, Input.Int 1)) else
						Input.Minus(e2, Input.Plus(e1, Input.Int 1)) 	(*** CHECK **)
		|   Neq -> if bit then Input.Minus(e1,e2) else Input.Minus(e2,e1) 
		|   _   -> raise (Failure "SupInf: getExpFromMsg BUG")
		else
			match l.op with
			Geq -> Input.Minus(e1, e2)
		|	Leq -> Input.Minus(e2, e1)
		|	Lt  -> Input.Minus(e2, Input.Plus(e1, Input.Int 1))
		|	Gt  -> Input.Minus(e1, Input.Plus(e2, Input.Int 1))
		|	Eq	-> if (bit) then Input.Minus(e1, e2) else Input.Minus(e2,e1)	
		|   Neq -> raise (Failure "NOT IMPLEMENTED")
		|   _   -> raise (Failure "SupInf: getExpFromMsg BUG")
		)	
	| Equality (a,b) -> 
		let e1 = expOfDag a
		and e2 = expOfDag b
		in
		if (msg.goal) then
			if bit then Input.Minus(e1, Input.Plus(e2, Input.Int 1))  
			else Input.Minus(e2, Input.Plus(e1, Input.Int 1))  
		else
			if bit then Input.Minus(e1,e2) else Input.Minus(e2,e1)




let expOfComb t = 
	let integer_t_list = List.rev (makeIntegerCoeffs t) in
	let il = (List.map (function i -> int_of_num i) integer_t_list)
	in
	let conslist = filterUndoListForMessage !undoStack in
	List.fold_right2 
		(function coeff -> function (msg,b) -> function exp ->
			if (coeff = 0) then exp else 
			Input.Plus (Input.Times ((Input.Int (coeff)), getExpFromMsg msg b), exp)) 
		il conslist (Input.Int 0)


let makeZeroComb t =
	let ilist = makeIntegerCoeffs t in
	let ex = List.filter (function a -> (a <>/ (Int 0))) (ilist) in
	List.fold_left (function exp -> function i -> 
			Input.Plus (Input.Times ((Input.Int (int_of_num i)), Input.Int 0), exp)) (Input.Int 0) ex

(** This is the main routine that generates proofs from the coefficient table **)
let generateProof t =
	let geq = makeCombProof t in	(* proof of geq *)
	let lt =  Proof.G ("geqCheck", [ E (makeZeroComb t); E (Input.Plus ((expOfComb t), Input.Int 1))]) in
	Proof.G ("geqLt_FI", [P geq; P lt])
 
let generateNoIntProof sT iT =
	let makeCoeffs x t =
		let integer_t_list = List.rev (makeIntegerCoeffs t) in
		let il = (List.map (function i -> int_of_num i) integer_t_list)
		in
		let conslist = filterUndoListForMessage !undoStack in
		List.fold_right2 
			(function coeff -> function (msg,b) -> function exp ->
				if (coeff = 0) then exp else 
				Input.Plus (Input.Times ((Input.Int (coeff)), getExpFromMsg msg b), exp)) 
			il conslist (Input.Int 0)
	in
	let getExp v =
		match v with
		  DebugNode s -> Input.Var(s)
		| RealNode nd -> expOfDag nd
	in
	let g = makeCombProof sT  (* proof of x >= inf *)
	and l = makeCombProof iT  (* proof of x<= sup *)
	in
	let (e1,e2) = (getEliminant sT, getEliminant iT) in
	let (exp2, fxp2) = (	(* Sup *)
		match e1 with
		  [Constant c; TimesVar(d, v)] -> if (d >=/ (Int 0)) then raise (Failure "SupInf generateNoIntProof: Bug 1")
										else (Input.Int (int_of_num c), Input.Times(getExp v, Input.Int (-(int_of_num d)) ))
		| [TimesVar(d, v); Constant c] -> if (d >=/ (Int 0)) then raise (Failure "SupInf generateNoIntProof: Bug 1")
										else (Input.Int (int_of_num c), Input.Times(getExp v, Input.Int (-(int_of_num d))))
		| _ -> raise (Failure "supInf: generateNoIntProof: sup coefficients do not eliminate all variables.")
	) and
	(exp1, fxp1) = (	(* Inf *)
		match e2 with
		  [Constant c; TimesVar(d, v)] -> if (d <=/ (Int 0)) then raise (Failure "SupInf generateNoIntProof: Bug 2")
										else (Input.Int (-(int_of_num c)), Input.Times(getExp v,Input.Int (int_of_num d)))
		| [TimesVar(d, v); Constant c] -> if (d <=/ (Int 0)) then raise (Failure "SupInf generateNoIntProof: Bug 2")
										else (Input.Int (-(int_of_num c)), Input.Times(getExp v,Input.Int (int_of_num d)))
		| _ -> raise (Failure "supInf: generateNoIntProof: inf coefficients do not eliminate all variables.")
	)
	in
	(* Inf *)
	let geq = Proof.G ("geq*i", [     I 1; I 1;
					  P (Proof.G ("eqCheck", [ 
						     E (Input.Plus(
                                                           Input.Times(Input.Int 1, Input.Minus(fxp1,exp1)), 
                                                                       Input.Minus(Input.Int 1, Input.Int 1))); 
						     E (Input.Minus(expOfComb iT,makeZeroComb iT)) ])) ;
						    P l]) and
	(* Sup *)
	leq = Proof.G ("geq*i", [                                 I 1; I 1;
								  P (Proof.G ("eqCheck", [ 
									E (Input.Plus(Input.Times(Input.Int 1, Input.Minus(exp2,fxp2)), Input.Minus(Input.Int 1, Input.Int 1))); 
									E (Input.Minus(expOfComb sT, makeZeroComb sT)) ])) ;
								  P g
								]) in
	Proof.G ("noInt_FI", [P geq; P leq])

(*** END  Proof generation *)



(* Check for contradiction *)
(* 
   Find the sup and the inf values for all variables in the new constraint under
   the set of constraints in undoStack. We assume that newConstraint has already been added to
   undoStack.
   
   Raise a contradiction if sup < inf or if the interval [inf, sup] contains no integers.
   
*)

let checkContra newConstraint = 
(* (ignore (VampyrePretty.printf "In checkContra")); *)
    let greater_than i s =
	  match (i,s) with
	  (Infinity, Infinity) -> true
	  | (NegInfinity, NegInfinity) -> true
	  | (Infinity, NegInfinity) -> true
	  | (NegInfinity, Infinity) -> false
	  | (Constant _, Infinity) -> false
	  | (NegInfinity, Constant _) -> false
	  | (Constant ci, Constant cs) -> (gt_num ci cs)
      | _ -> raise (Failure "checkContra: greater_than : Bug in sup-inf") 
    in
    let noIntInInterval(i,s) = (* [Check] *) 
	  match (i,s) with
	  (Infinity, Infinity) -> false	(* this will be caught earlier *)
	  | (NegInfinity, NegInfinity) -> false	(* this will be caught earlier *)
	  | (Infinity, NegInfinity) -> false	(* this will be caught earlier *)
	  | (NegInfinity, Infinity) -> false	(* this will be caught earlier *)
	  | (Constant _, Infinity) -> false
	  | (NegInfinity, Constant _) -> false
	  | (Constant ci, Constant cs) ->
	        (eq_num (floor_num ci) (floor_num cs)) && (eq_num (ceiling_num ci) (ceiling_num cs)) &&
            (not(is_integer_num ci) && not(is_integer_num cs))
      | _ -> raise (Failure "checkContra: noIntInInterval : Bug in sup-inf") 
	in
    let checkContraForVar v cons =
(*		(ignore (VampyrePretty.printf "in checkContraForVar")); *)

(*
	Sup and Inf return a list of size 1: the entry in the list is a TimesMin or TimesMax
	with exactly one expression
*)
		try
	    	let sval = List.hd (sup (Array.of_list (List.rev cons)) [TimesVar(one, v)])
			and ival = List.hd (inf (Array.of_list (List.rev cons)) [TimesVar(one, v)])
		in
		match (sval, ival) with
	(Infinity, NegInfinity) 
	| (TimesMin (Int 1, [(_,[Infinity])]), TimesMax(Int 1, [(_,[NegInfinity])]))  -> ()	(* we cannot claim anything :-( *)


	| (TimesMin (Int 1, [(sT,exs)]), NegInfinity) -> (
		(* maybe sup gives some indication of error *)
		let s = List.hd exs in
		let sTable = (getTable sT) in
		(** s = NegInfinity --> generate contradiction from s's table; **)
		match s with
		  NegInfinity -> (if(!debug) then ignore(VampyrePretty.printf "Raising contradiction: sup is NegInf");
						raise (Contradiction (generateProof (sTable))))
		| _ -> () (* give up -- cannot deduce anything *)


	)
	| (TimesMin (Int 1, [(sT,exs)]), TimesMax (Int 1, [(_,[NegInfinity])])) -> (
		(* maybe sup gives some indication of error *)
		let s = List.hd exs in
		let sTable = (getTable sT) in
		(** s = NegInfinity --> generate contradiction from s's table; **)
		match s with
		  NegInfinity -> (if(!debug) then ignore(VampyrePretty.printf "Raising contradiction: sup is NegInf");
				raise (Contradiction (generateProof (sTable))))

		| _ -> () (* give up -- cannot deduce anything *)

	  )


	| (Infinity, TimesMax (Int 1, [(iT,exi)])) -> (
		(* maybe inf gives some indication of error *)
		let i = List.hd exi in
		let iTable = (getTable iT) in
		(** i = Infinity --> generate contradiction from i's table; **)
		match i with
		  Infinity -> (if(!debug) then ignore(VampyrePretty.printf "Raising contradiction: inf is Inf");
					raise (Contradiction (generateProof (iTable))))
		| _ -> () (* give up -- cannot deduce anything *)

	)

	| (TimesMin (Int 1, [(_,[Infinity])]), TimesMax (Int 1, [(iT,exi)])) -> (
		(* maybe inf gives some indication of error *)
		let i = List.hd exi in
		let iTable = (getTable iT) in
		(** i = Infinity --> generate contradiction from i's table; **)
		match i with
		  Infinity -> (if(!debug) then ignore(VampyrePretty.printf "Raising contradiction: inf is Inf");
					raise (Contradiction (generateProof (iTable))))
		| _ -> () (* give up -- cannot deduce anything *)

	  )


	|	(TimesMin ((Int 1), [(sT, exs)]), TimesMax((Int 1), [(iT, exi)])) -> (
			let s = List.hd exs and i = List.hd exi in
			let sTable = (getTable sT) and iTable = (getTable iT) in
				(** s = NegInfinity --> generate contradiction from s's table; **)
		match s with
		  NegInfinity -> (if(!debug) then ignore(VampyrePretty.printf "Raising contradiction: sup is NegInf");
						raise (Contradiction (generateProof (sTable))))
		| _ -> (		
				(** i = Infinity --> generate contradiction from t's tablei **)
			match i with
				Infinity -> (if(!debug) then ignore(VampyrePretty.printf "Raising contradiction: inf is Inf");
						raise (Contradiction (generateProof iTable)))
				| _ -> (
				(**  else generate contradiction from both tables **)

		let t = addTables(sTable, iTable) in 
		if(greater_than i s)  then
			(ignore (if (!debug) then ignore(
				VampyrePretty.printf("SupInf: Found a contradiction : inf>sup.\n")));
			
			raise (Contradiction (generateProof t))
			)
        else (if (noIntInInterval(i,s)) then
				(ignore (if (!debug) then ignore(
    				VampyrePretty.printf("SupInf: Found a contradiction: No integer in interval.\n")));
				raise (Contradiction (generateNoIntProof sTable iTable)))
			  else ()
		     )
		)	
	)
	)
	(* sup / inf sometimes does not return min or max *)
	| _ -> raise (Failure "supInf: checkContra: sup / inf returns strange value!")

		(* Catch exception raised by sup / inf *)
		with SupInfException (tbloption, s) ->
			(if !debug then (ignore(VampyrePretty.printf "SupInf: Raising contradiction: Exception in Sup/Inf %s\n" s)));
			match tbloption with
			Some exTbl -> raise (Contradiction (generateProof (exTbl)))
			| None -> raise (Failure "No table sent back from supinf exception")
	
    in
	let rec checkContraForVarList el cons =
(*	(ignore (VampyrePretty.printf "In checkContraForVarList")); *)
	match el with
	  [] -> () 
	| v :: rest -> (checkContraForVar v cons); (checkContraForVarList rest cons)
	in
    let consList = List.map simp (filterUndoList !undoStack)	(* construct constraint list *)
    in
      let varList = (getVars newConstraint)       (* get a list of variables occurring in newConstraint *)
      in
	(if !debug then
	  (print_string "sup_inf: equations are:\n";
	   List.iter (function e -> print_string ((string_of_expr e) ^ " >= 0\n")) consList;
	   print_string "---\n");
	 match varList with
	   [] ->
	     (match newConstraint with
          [Constant c] ->
		 	if (ge_num c zero) then
		   		()
		 	else
		   		(ignore (if (!debug) then 
					  ignore (VampyrePretty.printf "SupInf: constant c < zero\n")) ;
		    	(ignore (if (!debug) then
		      	ignore (VampyrePretty.printf("SupInf: Found a contradiction.\n"))));
				let xp = 
					let ms = (findMessage (currentId()) !undoStack) in
					match ms with
					  Snapshot _ -> raise (Failure "SupInf: checkContra Bugg")
					| Constraint (a,b,c,_,_) ->
						Input.Plus(Input.Times(Input.Int 1, getExpFromMsg b c), Input.Int 0) in
		   	 	raise 
            	(Contradiction 
               		(Proof.G ("geqLt_FI", [ P (Proof.G ("geq*>=", [ I 1; P (getPf (currentId())); P (Proof.G ("geq0",[]))]));  (* the geq proof *)
                                       P (Proof.G ("geqCheck", [ Proof.E (Input.Plus(Input.Times(Input.Int 1,Input.Int 0), Input.Int 0)); 
																 Proof.E (Input.Plus (xp, Input.Int 1)) ]))
									 ])))) (* the Lt proof *)
                (* we have const >= 0 for const<0 *)
             | _  -> raise (Failure "sup_inf: checkContra: Strange newConstraint. Bug in simp?")
		  )
	 | _  -> (checkContraForVarList varList consList))

(* Convert a dag representation to an expr *)
(* throws NonlinearExpressionException *)
(* s is a Dag.node *)
exception NonlinearExpressionException of string

let rec makeExprFromDag s =
    match (s.op) with
      Plus  -> List.append (makeExprFromDag (List.nth s.args 0)) (makeExprFromDag (List.nth s.args 1))
    | Minus -> List.append (makeExprFromDag (List.nth s.args 0)) (negate (makeExprFromDag (List.nth s.args 1)))
    | Times ->
        let (e1,e2) = ((List.nth s.args 0), (List.nth s.args 1)) in 
          (match (e1.op, e2.op) with
            (Dag.Int k, ex) -> (List.map (mult_by (Int k)) (makeExprFromDag e2))
          | (ex, Dag.Int k) -> (List.map (mult_by (Int k)) (makeExprFromDag e1))
          | (ex1, ex2) -> 
              let (simpe1, simpe2) = (simp (makeExprFromDag e1),simp (makeExprFromDag e2)) 
              in 
                (match (simpe1, simpe2) with
                  ([Constant k], other) -> (List.map (mult_by k) other)
                | (other, [Constant k]) -> (List.map (mult_by k) other)
                | (_, _) -> raise (NonlinearExpressionException "SupInf: makeExprFromDag : Cannot handle nonlinear expression.")
                )
          )
    | Var i -> [ TimesVar(one, RealNode s) ]
    | Dag.Int i -> [ Constant(Int i) ]
	(* Note: Uninterpreted functions, floor, div, mod, ceil are all treated as a new variable *)
    | Fun f -> [ TimesVar(one, RealNode s) ]
    | Div _ -> [ TimesVar(one, RealNode s) ]
    | Mod _ -> [ TimesVar(one, RealNode s) ]
    | Floor _ -> [ TimesVar(one, RealNode s) ]
    | Ceil _ -> [ TimesVar(one, RealNode s) ]
    | _ -> raise (Failure "supInf.makeExprFromDag : Cannot handle these terms in expr")


let rec makeExprFromInput i =
    match i with
      Input.Int a -> [Constant (Int a)]
    | Input.Plus (a,b) -> List.append (makeExprFromInput a) (makeExprFromInput b)
    | Input.Minus (a,b) -> List.append (makeExprFromInput a) (negate (makeExprFromInput b))
    | Input.Times (e1,e2) ->
        (match (e1, e2) with
            (Input.Int k, ex) -> (List.map (mult_by (Int k)) (makeExprFromInput e2))
        | (ex, Input.Int k) -> (List.map (mult_by (Int k)) (makeExprFromInput e1))
        | (ex1, ex2) -> 
            let (simpe1, simpe2) = (simp (makeExprFromInput e1), simp (makeExprFromInput e2)) 
            in 
              (match (simpe1, simpe2) with
                ([Constant k], other) -> (List.map (mult_by k) other)
              | (other, [Constant k]) -> (List.map (mult_by k) other)
              | (_, _) -> raise (Failure "makeExprFromInput : Nonlinear expression?")))
    | Input.Var f ->
	let s = Dag.internExp i in
	  [ TimesVar(one, RealNode s) ]
	(* Note: Uninterpreted functions, floor, div, mod, ceil are all treated as a new variable *)
    | Input.Fun _ ->
	let s = Dag.internExp i in
	  [ TimesVar(one, RealNode s) ]
    | Input.Div _ ->
	let s = Dag.internExp i in
	  [ TimesVar(one, RealNode s) ]
    | Input.Mod _ ->
	let s = Dag.internExp i in
	  [ TimesVar(one, RealNode s) ]
    | Input.Floor _ ->
	let s = Dag.internExp i in
	  [ TimesVar(one, RealNode s) ]
    | Input.Ceil _ ->
	let s = Dag.internExp i in
	  [ TimesVar(one, RealNode s) ]

let simp_and_subtract_is_zero i1 i2 =
  let e1 = makeExprFromInput i1
  and e2 = makeExprFromInput i2
  in
    match simp (e1@(List.map (mult_by (Int (-1))) e2)) with
      [Constant zero] -> true
    | _ -> false

let simp_and_subtract_is_at_least_zero i1 i2 =
  let e1 = makeExprFromInput i1
  and e2 = makeExprFromInput i2
  in
    match simp (e1@(List.map (mult_by (Int (-1))) e2)) with
      [Constant c] -> ge_num c zero
    | _ -> false

let rec makeInputFromExpr e =
	let var v = 
		match v with
		RealNode s -> expOfDag s
		| DebugNode s -> raise (Failure "SupInf makeInputFromExpr : Variables should be dag nodes")
	in
	match e with
	[] -> raise (Failure "SupInf: makeInputFromExpr : Empty List!")
	| [a] -> (match a with
		Constant c -> Input.Int (int_of_num c)
	  | TimesVar(c, v) -> if (c =/ (Int 1)) then 
						(var v) else Input.Times(Input.Int (int_of_num c), var v)  
	  | _ -> raise (Failure "SupInf: makeInputFromExpr: These terms should not occur here")
	)
	| [a1; a2] -> (match (a1,a2) with 
			(Constant c, TimesVar(d,v)) -> 
										if (d >=/ (Int 0)) then Input.Plus(Input.Int (int_of_num c), 
									      if (d =/ (Int 1)) then (var v) else Input.Times(Input.Int (int_of_num d), var v))
										else Input.Minus(Input.Int (int_of_num c), 
                                          if (d =/ (Int 1)) then (var v) else Input.Times(Input.Int (-(int_of_num d)), var v))
			| (TimesVar(d,v), Constant c) ->
										if (d >=/ (Int 0)) then Input.Plus(Input.Int (int_of_num c), 
									      if (d =/ (Int 1)) then (var v) else Input.Times(Input.Int (int_of_num d), var v))
										else Input.Minus(Input.Int (int_of_num c), 
                                          if (d =/ (Int 1)) then (var v) else Input.Times(Input.Int (-(int_of_num d)), var v))
			| (TimesVar(d,v), TimesVar(c,w)) ->
						if(c >=/ (Int 0)) then Input.Plus(makeInputFromExpr [a1], makeInputFromExpr [a2]) else
							Input.Minus(makeInputFromExpr [a1], makeInputFromExpr [a2])
			| _ -> raise (Failure "SupInf: makeInputFromExpr: Strange argument!")
			
	)
	| a1::a2::rest -> let inp = makeInputFromExpr rest and lh = makeInputFromExpr [a1;a2] in
			Input.Plus(lh, inp)



let simplify i =
	let ei = simp (makeExprFromInput i) in
	makeInputFromExpr ei

	
(* Assert a new inequality. Add to the list of constraints (undoStack) and check for unsatisfiability *) 
let rec assume msg_list (msg: message) =
	ignore(
    if (!debug) then ignore
      (VampyrePretty.printf "SupInf : assume called by %s, and this is %s a goal\n" (msg.owner) (if msg.goal then "" else "not")));
  let assumeWorker msg node =	(* assumeWorker is the workhorse *)
	try	(* ignore nonlinear constraints *) 
    if (msg.goal) then (* need to negate expression *)
      match node.op with
	Geq ->
			 (* Convert to Expr2 - Expr1 -1 >= 0 *)
			 let con1 = makeExprFromDag (List.nth node.args 0) in
             let con2 = makeExprFromDag (List.nth node.args 1) in
             let newCon = (simp (con2 @ ((Constant (Int (-1)))::(negate con1)))) in
			 let diff_exp = (Input.Minus (expOfDag (List.nth node.args 1), Input.Plus(expOfDag (List.nth node.args 0), Input.Int 1))) in
			let temp_eq_pf = Proof.G ("eqCheck", 
			  [E (Input.Plus (Input.Times (Input.Int 1, Input.Minus (diff_exp, Input.Int 0)), 
			Input.Minus (Input.Int 1, Input.Int 1))); E diff_exp]) in
            let prf = (function () -> Proof.G ("geq*i", [I 1; I 1; P temp_eq_pf; P (Proof.G ("ltgeq", [P (!(Satproc.proveIt) msg)]))]))
             in
             undoStack := (Constraint (newCon, msg, false, prf, nextId())) :: !undoStack;
             checkContra (simp newCon)

	| Gt  -> (* Expr1 > Expr2 *)
			 (* Convert to Expr2 - Expr1 >= 0 *)
			 let con1 = makeExprFromDag (List.nth node.args 0) in
             let con2 = makeExprFromDag (List.nth node.args 1) in
             let newCon = con2 @ (negate con1) in
			 let diff_exp = Input.Minus (expOfDag (List.nth node.args 1), expOfDag (List.nth node.args 0)) in
			let temp_eq_pf = Proof.G ("eqCheck", 
			  [E (Input.Plus (Input.Times (Input.Int 1, Input.Minus (diff_exp, Input.Int 0)), 
			Input.Minus (Input.Int 1, Input.Int 1))); E diff_exp]) in
            let prf = (function () -> Proof.G ("geq*i", [I 1; I 1; P temp_eq_pf; P (Proof.G ("leqgeq", [P (!(Satproc.proveIt) msg)]))]))
             in
             undoStack := (Constraint (newCon, msg, false, prf, nextId())) :: !undoStack;
             checkContra (simp newCon)

	| Leq -> (* Expr1 <= Expr2 *) 
			(* Convert to Expr1 - Expr2 -1 >= 0 *)
			let con1 = makeExprFromDag (List.nth node.args 0) in
            let con2 = makeExprFromDag (List.nth node.args 1) in
            let newCon = (simp (con1 @ ((Constant (Int (-1)))::(negate con2)))) in
			let diff_exp = Input.Minus (expOfDag (List.nth node.args 0), Input.Plus(expOfDag (List.nth node.args 1), Input.Int 1)) in
			let temp_eq_pf = Proof.G ("eqCheck", 
			  [E (Input.Plus (Input.Times (Input.Int 1, Input.Minus (diff_exp, Input.Int 0)), 
			Input.Minus (Input.Int 1, Input.Int 1))); E diff_exp]) in
            let prf = (function () -> Proof.G ("geq*i", [I 1; I 1; P temp_eq_pf; P (Proof.G ("gtgeq", [P (!(Satproc.proveIt) msg)]))]))
			in
             undoStack := (Constraint (newCon, msg, false, prf, nextId())) :: !undoStack;
             checkContra (simp newCon)

	| Lt  -> (* Expr1 < Expr2 *)
			 (* Convert to Expr1 - Expr2 >= 0 *)
			 let con1 = makeExprFromDag (List.nth node.args 0) in
             let con2 = makeExprFromDag (List.nth node.args 1) in
             let newCon = (con1 @ (negate con2)) in
			 let diff_exp = Input.Minus (expOfDag (List.nth node.args 0), expOfDag (List.nth node.args 1)) in
			 let temp_eq_pf = Proof.G ("eqCheck", 
					[E (Input.Plus (Input.Times (Input.Int 1, Input.Minus (diff_exp, Input.Int 0)), 
						Input.Minus (Input.Int 1, Input.Int 1))); E diff_exp]) in
             let prf = (function () -> Proof.G ("geq*i", [I 1; I 1; P temp_eq_pf; P (!(Satproc.proveIt) msg)])) in
             undoStack := (Constraint (newCon, msg, false, prf, nextId())) :: !undoStack;
             checkContra (simp newCon)

	| Eq  ->	(* Expr1 = Expr2 *)
                (* Add two new constraints : 
                   Expr1-Expr2 -1 >=0 and
                   Expr2 - Expr1 -1 >=0
                   one by one
                *)

			 	(
                let con1 = makeExprFromDag (List.nth node.args 0) in
			 	let con2 = makeExprFromDag (List.nth node.args 1) in
				let newCon1 = con1 @ ((Constant (Int (-1))):: (negate con2)) in
				let newCon2 = con2 @ ((Constant (Int (-1))):: (negate con1)) in
			    let prf1 = (function () -> Input.Atom (Input.Geq (Input.Minus(expOfDag (List.nth node.args 0), 
									I.Plus (expOfDag (List.nth node.args 1), I.Int 1)),Input.Int 0)) ) 
			    and prf2 = (function () -> Input.Atom (Input.Geq (Input.Minus(expOfDag (List.nth node.args 1), 
									I.Plus (expOfDag (List.nth node.args 0), I.Int 1)),Input.Int 0)) ) 
				in
				let assumeVar1 = Input.newVar "S" and
				assumeVar2 = Input.newVar "S" in
                let id = nextId() in 
				(* first try with one *)
				try
				  undoStack := (Constraint (newCon1, msg, true, (function () -> Proof.Hyp assumeVar1), id)) :: !undoStack;
                  checkContra (simp newCon1)
				with (Contradiction c) ->
					try
                  	(* then try the other one *) 
				  		undoStack := ((Constraint (newCon2, msg, false,(function () -> Proof.Hyp assumeVar2), id)):: (List.tl !undoStack));
                  		checkContra (simp newCon2)
					with (Contradiction d) -> (* generate proof *)
						let firstPart = Proof.Impi(assumeVar1, prf1(), c) and
						secondPart = Proof.Impi(assumeVar2, prf2(), d) 
						in
						raise (Contradiction (Proof.G("pathSplit_FI", [ 
									P ( (!Satproc.proveIt) msg);
									P firstPart;
									P secondPart
								]
								)))
					
                 ) 
	|   Neq ->  (* Expr1 != Expr2 *) 
				(* Add two constraints and check for contradiction *)
			 	let con1 = makeExprFromDag (List.nth node.args 0) in
			 	let con2 = makeExprFromDag (List.nth node.args 1) in
				let newCon1 = con1 @ (negate con2) in
				let newCon2 = con2 @ (negate con1) in
				let (e0,e1) = (expOfDag (List.nth node.args 0), expOfDag (List.nth node.args 1)) in 
				let diff_exp1 = Input.Minus (e0, e1) in
				let diff_exp2 = Input.Minus (e1, e0) in
				let tmp_eq_pf1 = Proof.G ("eqCheck",
               [E (Input.Plus (Input.Times (Input.Int 1, Input.Minus (diff_exp1, Input.Int 0)),
               Input.Minus (Input.Int 1, Input.Int 1))); E diff_exp1]) in
				let tmp_eq_pf2 = Proof.G ("eqCheck",
               [E (Input.Plus (Input.Times (Input.Int 1, Input.Minus (diff_exp2, Input.Int 0)),
               Input.Minus (Input.Int 1, Input.Int 1))); E diff_exp2]) in
				let prf1 = (function () -> Proof.G ("geq*i", [ I 1; I 1; P tmp_eq_pf1; P (Proof.G ("eqgeq", [P (!(Satproc.proveIt) msg)]))])) in 
				let prf2 = (function () -> Proof.G ("geq*i", [ I 1; I 1; P tmp_eq_pf2; P (Proof.G ("eqgeqr", [P (!(Satproc.proveIt) msg)]))])) in 
				undoStack := (Constraint (newCon1, msg, true, prf1, nextId())) :: 
						     ((Constraint (newCon2, msg, false, prf2, nextId())):: !undoStack);
                checkContra (simp newCon1)	(* all variables in newCon2 are also in newCon1 *)
				
	| _   -> ()	(* ignore *)
    
    else (* goal = false *)
    match node.op with
	  Geq ->
             if ((List.nth node.args 1).op == (Dag.Int 0)) then  (* Expr >= 0 *)
			 	let newCon = makeExprFromDag (List.nth node.args 0) in
			    let prf = (function () -> (!(Satproc.proveIt) msg)) in
			 	undoStack := (Constraint (newCon, msg, false, prf, nextId())) :: !undoStack;
				checkContra (simp newCon )
			 else (* Expr1 >= Expr2 *)
			 	let con1 = makeExprFromDag (List.nth node.args 0) in
			 	let con2 = makeExprFromDag (List.nth node.args 1) in
				let newCon = con1 @ (negate con2) in
				let diff_exp = Input.Minus (expOfDag (List.nth node.args 0), expOfDag (List.nth node.args 1)) in
				let temp_eq_pf = Proof.G ("eqCheck", 
					[E (Input.Plus (Input.Times (Input.Int 1, Input.Minus (diff_exp, Input.Int 0)), 
						Input.Minus (Input.Int 1, Input.Int 1))); E diff_exp]) in
                let prf = (function () -> Proof.G ("geq*i", [I 1; I 1; P temp_eq_pf; P (!(Satproc.proveIt) msg)])) in
				undoStack := (Constraint (newCon, msg, false, prf, nextId())) :: !undoStack;
                checkContra (simp newCon)

	| Gt  -> (* Expr1 > Expr2 *)
			 (* Convert to Expr1 - Expr2 -1 >= 0 *)
			 let con1 = makeExprFromDag (List.nth node.args 0) in
             let con2 = makeExprFromDag (List.nth node.args 1) in
             let newCon = con1 @ ((Constant (Int (-1)))::(negate con2)) in
			 let diff_exp = Input.Minus (expOfDag (List.nth node.args 0), Input.Plus(expOfDag (List.nth node.args 1), Input.Int 1)) in
             let temp_eq_pf = Proof.G ("eqCheck", 
               [E (Input.Plus (Input.Times (Input.Int 1, Input.Minus (diff_exp, Input.Int 0)), 
               Input.Minus (Input.Int 1, Input.Int 1))); E diff_exp]) in
			 let prf = (function () -> Proof.G ("geq*i", [I 1; I 1; P temp_eq_pf; P (Proof.G ("gtgeq", [P (!(Satproc.proveIt) msg)]))])) in
             undoStack := (Constraint (newCon, msg, false, prf, nextId())) :: !undoStack;
             checkContra (simp newCon)

	| Leq -> 
             if ((List.nth node.args 0).op == (Dag.Int 0)) then  (* Expr <= 0 *)
                let newCon = makeExprFromDag (List.nth node.args 1) in
	            let diff_exp = Input.Minus (expOfDag (List.nth node.args 1), Input.Int 0 ) in
                let temp_eq_pf = Proof.G ("eqCheck", 
                    [E (Input.Plus (Input.Times (Input.Int 1, Input.Minus (diff_exp, Input.Int 0)), 
                        Input.Minus (Input.Int 1, Input.Int 1))); E diff_exp]) in
                let prf = (function () -> Proof.G ("geq*i", [I 1; I 1; P temp_eq_pf; P (Proof.G ("leqgeq", [P (!(Satproc.proveIt) msg)]))])) in
                undoStack := (Constraint (newCon, msg, false, prf, nextId())) :: !undoStack;
                checkContra (simp newCon)
             else (* Expr1 <= Expr2 *)
                let con1 = makeExprFromDag (List.nth node.args 0) in
                let con2 = makeExprFromDag (List.nth node.args 1) in
                let newCon = con2 @ (negate con1) in
	            let diff_exp = Input.Minus (expOfDag (List.nth node.args 1), expOfDag (List.nth node.args 0)) in
                let temp_eq_pf = Proof.G ("eqCheck", 
                    [E (Input.Plus (Input.Times (Input.Int 1, Input.Minus (diff_exp, Input.Int 0)), 
                        Input.Minus (Input.Int 1, Input.Int 1))); E diff_exp]) in
                let prf = (function () -> Proof.G ("geq*i", [I 1; I 1; P temp_eq_pf; P (Proof.G ("leqgeq", [P (!(Satproc.proveIt) msg)]))])) in
                undoStack := (Constraint (newCon, msg, false, prf, nextId())) :: !undoStack;
                checkContra (simp newCon)

	| Lt  -> (* Expr1 < Expr2 *)
			 (* Convert to Expr2 - Expr1 -1 >= 0 *)
			 let con1 = makeExprFromDag (List.nth node.args 0) in
             let con2 = makeExprFromDag (List.nth node.args 1) in
             let newCon = con2 @ ((Constant (Int (-1)))::(negate con1)) in
			 let diff_exp = Input.Minus (expOfDag (List.nth node.args 1), Input.Plus(expOfDag (List.nth node.args 0), Input.Int 1)) in
             let temp_eq_pf = Proof.G ("eqCheck", 
               [E (Input.Plus (Input.Times (Input.Int 1, Input.Minus (diff_exp, Input.Int 0)),
               Input.Minus (Input.Int 1, Input.Int 1))); E diff_exp]) in
             let prf = (function () -> Proof.G ("geq*i", [I 1; I 1; P temp_eq_pf; P (Proof.G ("ltgeq", [P (!(Satproc.proveIt) msg)]))])) in
             undoStack := (Constraint (newCon, msg, false, prf, nextId())) :: !undoStack;
             checkContra (simp newCon)

	| Eq  ->	(* Expr1 = Expr2 *)
                (* Add two new constraints : 
                   Expr1-Expr2>=0 and
                   Expr2 - Expr1 >=0
                *)
				
				(* ignore equalities generated by simplex : I already "know" them implicitly *)

				if (msg.owner = "SX") then () 

				(** The above is a hideous hack, but I cannot think of any way of doing this better. **)
				(** My only saving point is that the procedure is correct even without the above hack :-) **)
				(** It is SLOW, but ... **)


				else ( 
			 	let con1 = makeExprFromDag (List.nth node.args 0) in
			 	let con2 = makeExprFromDag (List.nth node.args 1) in
				let newCon1 = con1 @ (negate con2) in
				let newCon2 = con2 @ (negate con1) in
				let (e0,e1) = (expOfDag (List.nth node.args 0), expOfDag (List.nth node.args 1)) in 
				let diff_exp1 = Input.Minus (e0, e1) in
				let diff_exp2 = Input.Minus (e1, e0) in
				let tmp_eq_pf1 = Proof.G ("eqCheck",
               [E (Input.Plus (Input.Times (Input.Int 1, Input.Minus (diff_exp1,Input.Int 0)), 
               Input.Minus (Input.Int 1, Input.Int 1))); E diff_exp1]) in
				let tmp_eq_pf2 = Proof.G ("eqCheck",
               [E (Input.Plus (Input.Times (Input.Int 1, Input.Minus (diff_exp2, Input.Int 0)),
               Input.Minus (Input.Int 1, Input.Int 1))); E diff_exp2]) in
				let prf1 = (function () -> Proof.G ("geq*i", [ I 1; I 1; P tmp_eq_pf1; P (Proof.G ("eqgeq", [P (!(Satproc.proveIt) msg)]))])) in 
				let prf2 = (function () -> Proof.G ("geq*i", [ I 1; I 1; P tmp_eq_pf2; P (Proof.G ("eqgeqr", [P (!(Satproc.proveIt) msg)]))])) in 
				undoStack := (Constraint (newCon1, msg, true, prf1, nextId())) :: 
						     ((Constraint (newCon2, msg, false, prf2, nextId())):: !undoStack);
                checkContra (simp newCon1)	(* all variables in newCon2 are also in newCon1 *)
				)

	|   Neq ->  (* Expr1 != Expr2 *)
				(* raise (Failure "NOT IMPLEMENTED") *)
				()	(* ignore : this is out of your league :-) *)



 
	| _   -> ()	(* ignore *)

	(* Ignore nonlinear expressions *)
	with (NonlinearExpressionException warnMsg) -> if !debug then ignore(VampyreErrormsg.warn "%s\n" warnMsg)
  in
  ignore (
    (match msg.body with
      Literal (l) -> ignore (assumeWorker msg l)
    | Equality (a,b) -> if ((Dag.isExp a) && (Dag.isExp b)) then
						    let node = (newNode Eq [a;b]) in
						    ignore (assumeWorker msg node)
		                else () 
    )
    ); (msg_list: message list )  (* return list of equalities *)
	(* we do not yield any new messages, since we do not generate inequalities *)


(* Since we are not generating any equalities, proveIt for us is trivial:
   no one should ever call us.
*)
let proveIt _ = VampyreErrormsg.s (VampyreErrormsg.bug "SupInf: SupInf does not generate equalities. So how come proveIt is being called?")

(* 
    Initialize the decision procedure and create a table of the entry points 
*)
let initialize id =
    supInfId := id;
    {
        Satproc.snapshot = snapshot;
        Satproc.undo = undo;
        Satproc.assume = assume;
        Satproc.proveIt = proveIt;
        Satproc.debug = debug;
        Satproc.ident = id;
	Satproc.reinitialize = function () -> ();
     }
(*
	Re-initialize the decision procedure.
*)
let reinitialize () =
	resetId (-1);
	undoStack := []

(* regression testing: *)
let negone = Int (-1)
let two = Int 2
let three = Int 3

exception RegressionFailure of expr * string

let same_tables t1 t2 =
  match (t1, t2) with
    (None, None) -> true
  | (Some _, None) -> false
  | (None, Some _) -> false
  | (Some tbl1, Some tbl2) ->
      (print_string ("damn: " ^ (string_of_expr [TimesMin (one, [(Some tbl1,[])])]) ^ (string_of_expr [TimesMin (one, [(Some tbl2,[])])]));
       tbl1 = tbl2)

let regression = ref []
let addreg el = regression := el::(!regression)
let check_regressions () =
  let rec helper = function
      (eqns, supinf, query, (constants, answer), testname)::rest ->
	(if (!debug) then ignore (print_string ("SupInf: Running regression #" ^ testname ^ "\n")));
	let return = supinf eqns query in
	  (match return with
	    [TimesMin (one, [ (table, term) ])] ->
	      ((*if (same_tables table constants) then
		raise (RegressionFailure (return, "table constants do not agree in test a#" ^ testname));*)
	       match term with
		 [Constant a] ->
		   (match answer with
		     [Constant b] ->
		       if eq_num a b then
			 helper rest
		       else
			 raise (RegressionFailure (return, "regression failed on test #" ^ testname))
		   | _ -> raise (RegressionFailure (return, "regression failed on test #" ^ testname)))
	       | [Infinity] ->
		   (match answer with
		     [Infinity] -> helper rest
		   | _ -> raise (RegressionFailure (return, "regression failed on test #" ^ testname)))
	       | [NegInfinity] ->
		   (match answer with
		     [NegInfinity] -> helper rest
		   | _ -> raise (RegressionFailure (return, "regression failed on test #" ^ testname)))
	       | _ ->
		   raise (RegressionFailure (return, "non-constant value returned in test #" ^ testname)))
	  | [TimesMax (one, [ (table, term) ])] ->
	      ((*if (same_tables table constants) then
		 raise (RegressionFailure (return, "table constants do not agree in test b#" ^ testname))
	      else
		();*)
	       match term with
		 [Constant a] ->
		   (match answer with
		     [Constant b] ->
		       if eq_num a b then
			 helper rest
		       else
			 raise (RegressionFailure (return, "regression failed on test #" ^ testname))
		   | _ -> raise (RegressionFailure (return, "regression failed on test #" ^ testname)))
	       | [Infinity] ->
		   (match answer with
		     [Infinity] -> helper rest
		   | _ -> raise (RegressionFailure (return, "regression failed on test #" ^ testname)))
	       | [NegInfinity] ->
		   (match answer with
		     [NegInfinity] -> helper rest
		   | _ -> raise (RegressionFailure (return, "regression failed on test #" ^ testname)))
	       | _ ->
		   raise (RegressionFailure (return, "non-constant value returned in test #" ^ testname)))
	  | _ -> raise (RegressionFailure (return, "strange value returned by sup or inf in test #" ^ testname)))
    | [] -> ()
  in
    helper (!regression)

(* the example from Shostak's paper: *)
let eqns = [ [TimesVar(one, DebugNode "y")];
	     [TimesVar(one, DebugNode "y"); TimesVar(negone, DebugNode "x")];
	     [Constant one; TimesVar (negone, DebugNode "y"); TimesVar (negone, DebugNode "x")]]
let supinf = sup
let query = [TimesVar (one, DebugNode "x")]
let answer = (Some [|zero; one; one|], [Constant (div_num one two)])
let testname = "1"
let _ = addreg (Array.of_list eqns, supinf, query, answer, testname)

let supinf = inf
let query = [TimesVar (one, DebugNode "x")]
let answer = (None, [NegInfinity])
let testname = "2"
let _ = addreg (Array.of_list eqns, supinf, query, answer, testname)

let supinf = sup
let query = [TimesVar (one, DebugNode "y")]
let answer = (Some [|zero; zero; one|], [Infinity])  (* there's really non proof needed at all, but this is what it returns. *)
let testname = "3"
let _ = addreg (Array.of_list eqns, supinf, query, answer, testname)

let supinf = inf
let query = [TimesVar (one, DebugNode "y")]
let answer = (Some [|one; zero; zero|], [Constant zero])
let testname = "4"
let _ = addreg (Array.of_list eqns, supinf, query, answer, testname)

let eqns = [ [TimesVar (Int 3, DebugNode "x"); Constant negone];
	     [TimesVar (one, DebugNode "y")];
	     [TimesVar (one, DebugNode "y"); TimesVar (negone, DebugNode "x")];
	     [Constant one; TimesVar (negone, DebugNode "y"); TimesVar (negone, DebugNode "x")] ]
let supinf = sup
let query = [TimesVar (one, DebugNode "x")]
let answer = (Some [|zero; zero; one; one|], [Constant (div_num one two)])
let testname = "5"
let _ = addreg (Array.of_list eqns, supinf, query, answer, testname)

let supinf = inf
let answer = (Some [|div_num one (Int 3); zero; zero; zero|], [Constant (div_num one (Int 3))])
let testname = "6"
let _ = addreg (Array.of_list eqns, supinf, query, answer, testname)

let supinf = sup
let query = [TimesVar (one, DebugNode "y")]
let answer = (Some [|div_num one (Int 3); zero; zero; one|], [Constant (div_num two (Int 3))])
let testname = "7"
let _ = addreg (Array.of_list eqns, supinf, query, answer, testname)

let supinf = inf
let answer = (Some [|div_num one (Int 3); zero; one; zero|], [Constant (div_num one (Int 3))])
let testname = "8"
let _ = addreg (Array.of_list eqns, supinf, query, answer, testname)

let eqns = [ [TimesVar (one, DebugNode "x"); TimesVar (negone, DebugNode "y")];
	     [Constant two; TimesVar (negone, DebugNode "y"); TimesVar (negone, DebugNode "x")];
	     [Constant (Int (-3)); TimesVar (one, DebugNode "x")];
	     [Constant one; TimesVar (one, DebugNode "y")]
	   ]

let supinf = inf
let query = [TimesVar (one, DebugNode "y")]
let answer = (Some [|zero; zero; zero; one|], [Constant (Int (-1))])
let testname = "9"
let _ = addreg (Array.of_list eqns, supinf, query, answer, testname)

let supinf = sup
let answer = (Some [|zero; one; one; zero|], [Constant (Int (-1))])
let testname = "10"
let _ = addreg (Array.of_list eqns, supinf, query, answer, testname)

let supinf = inf
let query = [TimesVar (one, DebugNode "x")]
let answer = (Some [|zero; zero; one; zero|], [Constant (Int 3)])
let testname = "11"
let _ = addreg (Array.of_list eqns, supinf, query, answer, testname)

let supinf = sup
let answer = (Some [| zero;one;zero;one |], [Constant (Int 3)])
let testname = "12"
let _ = addreg (Array.of_list eqns, supinf, query, answer, testname)

let eqns = [ [TimesVar (one, DebugNode "x"); TimesVar (one, DebugNode "z"); TimesVar (negone, DebugNode "y")];
	     [TimesVar (one, DebugNode "y"); TimesVar (one, DebugNode "z"); TimesVar (negone, DebugNode "x")] ]
let supinf = inf
let query = [TimesVar (one, DebugNode "z")]
let answer = (Some [|one; one|], [Constant (Int 0)])
let testname = "13"
let _ = addreg (Array.of_list eqns, supinf, query, answer, testname)
