(*
 * Description:
 *
 * Copyright (c) 2000 by
 *  David Blei			blei@cs.berkeley.edu
 *  Chris Harrelson		chrishtr@cs.berkeley.edu
 *  Ranjit Jhala		jhala@cs.berkeley.edu
 *  Rupak Majumdar		rupak@cs.berkeley.edu
 *  George C. Necula	necula@cs.berkeley.edu
 *  Shree P. Rahul		sprahul@cs.berkeley.edu
 * 	Wes Weimer			weimer@cs.berkeley.edu
 *  Dror Weitz			dror@cs.berkeley.edu
 *   
 * All rights reserved.  Permission to use, copy, modify and distribute
 * this software for research purposes only is hereby granted, 
 * provided that the following conditions are met: 
 * 1.  Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 * 3. The name of the authors may not be used to endorse or promote products derived from 
 * this software without specific prior written permission. 
 *
 * DISCLAIMER:
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR IMPLIED 
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF 
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *)
(* S. P. Rahul
 * 
 * Satisfiability procedure for typing and memory safety predicates
 * 
 * The predicates this module deals with are typing i.e (of ...) and safety 
 * i.e. (saferd4 ...) and (safewr4 ...). 
*)

module E = VampyreErrormsg
module I = Input
module P = Proof
module S = Satproc
module SG = Subgoal
module D = Dag
open VampyrePretty


let debug = ref false   (* no idea how this gets the value true *) 				        
let typId = ref ""

exception Fail

(*----------------------------------------------------------------------*)
(* undoStack manipulations*)

type undoElem = 
    Snapshot  (* marker *)
  | Assume of  S.message  

let undoStack : undoElem list ref = ref []

let snapshot () = 
	undoStack := Snapshot :: !undoStack

let undo () =
  let rec loop = function
      [] -> E.s (E.bug "TYP tried to to pop an empty stack\n")
    | Snapshot :: rest -> undoStack := rest
    | _ :: rest -> loop rest 
  in
  loop !undoStack

let push x = 
  undoStack :=  x :: !undoStack

(*----------------------------------------------------------------------*)
(* Since TYP will never broadcast any assertions (it's job is to prove atoms) 
 * proveIt should never be called *)
let proveIt msg = E.s (E.unimp "TYP.proveIt not implemented (does not need to be)\n")

(* Determines if the assume we received is (of .... ..) *)
let typingAss body  =
  match body with
    S.Equality (_,_) -> false 
  | S.Literal x -> 
      (match x.D.op with
        D.BFun "of" -> true
      | _ -> false)
  

(*----------------------------------------------------------------------*)
(* Takes a dag node which is meant to represent a type and returns a tuple. 
 * The first element in the tuple is the size in bytea and the second is a 
 * proof of this fact *)

(* This only works for things that are 4 bytes because that's what we have 
 * axioms for *)

let getSize (t:I.exp) = 
  match t with
    I.Fun("int",[]) -> (4, P.G("szint",[]))
  | I.Fun("bool",[]) -> (4, P.G("szbool",[]))
  | I.Fun("array",[ro;t;len]) -> (4, P.G("szarr",[P.E ro; P.E t; P.E len]))
  | I.Fun("ptropt",[t;len]) -> (4, P.G("ptropt",[P.E t;P.E len]))
  | _ -> E.s (E.unimp "TYP: getSize: can't (don't need to) handle structured types")

(* Return only the size. This works for any kind of structured type, 
 * therefore it does return the proof *)

let rec getSizeOnly (t:I.exp) =
  match t with
    I.Fun("int",[]) -> 4
  | I.Fun("bool",[]) -> 4
  | I.Fun("array",[ro;t;len]) -> 4
  | I.Fun("ptropt",[t;len]) -> 4
  | I.Fun("tuple2",[t1;t2]) -> getSizeOnly t1 + getSizeOnly t2
  | I.Fun("tuple3",[t1;t2;t3]) -> getSizeOnly t1 + getSizeOnly t2 + getSizeOnly t3
  | I.Fun("tuple4",[t1;t2;t3;t4]) -> getSizeOnly t1 + getSizeOnly t2 + getSizeOnly t3 + getSizeOnly t4


(* Given a (possibly structured type) t and an integer offset i, returns a 
 * proof that there is something sensible (i.e. of valid type) at that offset 
 * Elements of tuples have to be atomic (4 bytes) because that's what we can 
 * prove with the given axioms *)
(* Also returns the type of stuff at that offset. Useful for safewr *)

let proveFieldAccess (t:I.exp) (i:int)  = 

  match (t,i) with 
    (I.Fun("int",[]),0)  -> (P.G("fo10",[P.P(P.G("szint",[]))]),I.Fun("int",[]))
  | (I.Fun("int",[]),_)  -> E.s(E.error "TYP: fieldPrf 1")
  | (I.Fun("bool",[]),0) -> (P.G("fo10",[P.P(P.G("szbool",[]))]), I.Fun("bool",[]))
  | (I.Fun("bool",[]),_) ->  E.s(E.error "TYP: fieldPrf 2")
  | (I.Fun("array",[ro;t;len]),0) -> 
      (P.G("fo10",[P.P(P.G("szarr",[P.E ro; P.E t; P.E len]))]), I.Fun("array",[]))

  | (I.Fun("array",[ro;t;len]),_) -> E.s(E.error "TYP: fieldPrf 3")
  | (I.Fun("ptropt",[t;len]),0) -> 
      (P.G("fo10",[P.P(P.G("szptropt",[P.E t; P.E len]))]),I.Fun("array",[]))
  | (I.Fun("ptropt",[t;len]),_) ->E.s(E.error "TYP: fieldPrf 4")
        
  | (I.Fun("tuple2",[t1;t2]),0) ->
      (match (getSize t1, getSize t2) with
        ((4,t1Prf),(4,t2Prf)) -> (P.G("fo20",[P.P t1Prf; P.P t2Prf]), t1)
      | _ -> E.s(E.error "TYP: fieldPrf: tuple of structured type perhaps?"))

  | (I.Fun("tuple2",[t1;t2]),4) ->
      (match (getSize t1, getSize t2) with
        ((4,t1Prf),(4,t2Prf)) -> (P.G("fo21",[P.P t1Prf; P.P t2Prf]),t2)
      | _ -> E.s(E.error "TYP: fieldPrf: tuple of structured type perhaps?"))

  | (I.Fun("tuple3",[t1;t2;t3]),0) -> 
      (match (getSize t1, getSize t2, getSize t3) with
        ((4,t1Prf),(4,t2Prf),(4,t3Prf)) -> (P.G("fo30",[P.P t1Prf; P.P t2Prf; P.P t3Prf]),t1)
      | _ -> E.s(E.error "TYP: fieldPrf: tuple of structured type perhaps?"))

  | (I.Fun("tuple3",[t1;t2;t3]),4) -> 
      (match (getSize t1, getSize t2, getSize t3) with
        ((4,t1Prf),(4,t2Prf),(4,t3Prf)) -> (P.G("fo31",[P.P t1Prf; P.P t2Prf; P.P t3Prf]),t2)
      | _ -> E.s(E.error "TYP: fieldPrf: tuple of structured type perhaps?"))

  | (I.Fun("tuple3",[t1;t2;t3]),8) -> 
      (match (getSize t1, getSize t2, getSize t3) with
        ((4,t1Prf),(4,t2Prf),(4,t3Prf)) -> (P.G("fo32",[P.P t1Prf; P.P t2Prf; P.P t3Prf]),t3)
      | _ -> E.s(E.error "TYP: fieldPrf: tuple of structured type perhaps?"))

  | (I.Fun("tuple4",[t1;t2;t3;t4]),0) -> 
      (match (getSize t1, getSize t2, getSize t3, getSize t4) with
        ((4,t1Prf),(4,t2Prf),(4,t3Prf),(4,t4Prf)) -> (P.G("fo40",[P.P t1Prf; P.P t2Prf; P.P t3Prf;
                                                                  P.P t4Prf]),t1)
      | _ -> E.s(E.error "TYP: fieldPrf: tuple of structured type perhaps?"))

  | (I.Fun("tuple4",[t1;t2;t3;t4]),4) -> 
      (match (getSize t1, getSize t2, getSize t3, getSize t4) with
        ((4,t1Prf),(4,t2Prf),(4,t3Prf),(4,t4Prf)) -> (P.G("fo41",[P.P t1Prf; P.P t2Prf; P.P t3Prf; 
                                                                  P.P t4Prf]),t2)
      | _ -> E.s(E.error "TYP: fieldPrf: tuple of structured type perhaps?"))

  | (I.Fun("tuple4",[t1;t2;t3;t4]),8) -> 
      (match (getSize t1, getSize t2, getSize t3, getSize t4) with
        ((4,t1Prf),(4,t2Prf),(4,t3Prf),(4,t4Prf)) -> (P.G("fo42",[P.P t1Prf; P.P t2Prf; P.P t3Prf; 
                                                                  P.P t4Prf]),t3)
      | _ -> E.s(E.error "TYP: fieldPrf: tuple of structured type perhaps?"))

  | (I.Fun("tuple4",[t1;t2;t3;t4]),12) -> 
      (match (getSize t1, getSize t2, getSize t3, getSize t4) with
        ((4,t1Prf),(4,t2Prf),(4,t3Prf),(4,t4Prf)) -> (P.G("fo43",[P.P t1Prf; P.P t2Prf; P.P t3Prf; 
                                                                  P.P t4Prf]),t4)
      | _ -> E.s(E.error "TYP: fieldPrf: tuple of structured type perhaps?"))
  | _ -> E.s(E.error "TYP: fieldPrf: unknown type")
      

(*================================================================================*)
(* Decompose ADDR into A + I. If ADDR does not contain a Plus, then I is zero *)
(* ADDR is a D.node *)
(* returnns (A,I) where both A is D.node and I is I.exp *)

let decomposeAddr (addr:D.node) = 
  if (!debug)
  then ignore (printf "\ntrying to decompose addr %a\n" D.dn addr);

  match addr with
   | {D.op = D.Plus; D.args = [{D.op = D.Plus; D.args = [a1;a2]};a3]} ->
      let D.DExp a2Val = D.nodeVal a2 in
      let D.DExp a3Val = D.nodeVal a3 in
      (a1,I.Plus(a2Val,a3Val))
   | {D.op = D.Minus; D.args = [{D.op = D.Plus; D.args = [a1;a2]};a3]} ->
       let D.DExp a2Val = D.nodeVal a2 in
       let D.DExp a3Val = D.nodeVal a3 in
       (a1, I.Minus(a2Val,a3Val))
      
  | {D.op = D.Plus; D.args = [a;i]} -> 
      let D.DExp i_inp = D.nodeVal i in
      (a,i_inp)
  | {D.op = D.Fun "sel4"} -> (addr,I.Int(0))
  | {D.op = D.Var (_)} -> (addr,I.Int(0)) (* can be simply a variable *)
  | {D.op = D.Fun (_);D.args = []} -> (addr, I.Int(0)) (* can be a global var *)
  | _ -> raise Fail

(* Decompose E, the index into an array. E = SZ*E' + OFFSET *)
(* SZ:int, E:I.exp *)
(* returns (E',OFFSET) where both are I.exp *)

let rec decomposeIndex (e:I.exp) (sz:int) =
  if (!debug)
  then ignore (printf "\ntrying to decompose index %a\n" I.de e);
  
  match e with
    I.Plus(I.Int 0, x) -> decomposeIndex x sz
  | I.Plus(I.Plus(I.Int 0,x),y) -> decomposeIndex (I.Plus(x,y)) sz
  | _ -> 
      let (e',i) = 
        (match e with
          I.Minus(I.Times(e',I.Int(x)),I.Int(y)) when x=sz -> 
            if (y mod x = 0)
            then (I.Minus(e',I.Int(y/x)),0)
            else E.s(E.unimp "TYP:decomposeIndex")
        | I.Plus(I.Times(I.Int sz',e'),I.Int i) when sz=sz' ->(e',i)
        | I.Plus(I.Times(e',I.Int sz'),I.Int i) when sz=sz' -> (e',i)
        | I.Plus(I.Int i, I.Times(I.Int sz',e')) when sz=sz' -> (e',i)
        | I.Plus(I.Int i, I.Times(e',I.Int sz')) when sz=sz' -> (e',i)
        | I.Times(I.Int sz',e') -> (e',0)
        | I.Times(e',I.Int sz') -> (e',0)
        | I.Int idx ->
            (I.Int(idx/sz), (idx mod sz))
        | I.Times(I.Minus(I.Times(x,I.Int n1),y),I.Int n2) when x=y && (n1-1)*n2=sz ->
            (x,0)
        | I.Plus(I.Times(I.Minus(I.Times(x,I.Int n1),y),I.Int n2),I.Int n3) when x=y && (n1-1)*n2=sz->
            (x,n3)
        | _ -> raise Fail)
      in
      if (i >= sz)
      then 
        (let e' = I.Plus(e',I.Int(i/sz)) in
        let i  = i mod sz in
        (I.Plus(I.Times(e',I.Int sz),I.Int i),e',i))
      else 
        (e,e',i)
      
(*----------------------------------------------------------------------*)
(* prove typing predicates (of e1 e2) *)
(* msg is the original message (verified to be of the above form)             
 * args is [e1;e2] *)
  
let rec processTypingGoal msg args  =
  match args with
    [{D.op = (D.Int i)}; {D.op = (D.Fun "exn"); D.args = []}] -> 
      (match i with
        0 -> raise (S.Contradiction (P.G("not_FI",[P.P(P.G("ofExn0",[]));P.P(!S.proveIt msg)])))
      | 1 -> raise (S.Contradiction (P.G("not_FI",[P.P(P.G("ofExn1",[]));P.P(!S.proveIt msg)])))
      | 2 -> raise (S.Contradiction (P.G("not_FI",[P.P(P.G("ofExn2",[]));P.P(!S.proveIt msg)])))
      | 3 -> raise (S.Contradiction (P.G("not_FI",[P.P(P.G("ofExn3",[]));P.P(!S.proveIt msg)])))
      | 4 -> raise (S.Contradiction (P.G("not_FI",[P.P(P.G("ofExn4",[]));P.P(!S.proveIt msg)])))
      | 5 -> raise (S.Contradiction (P.G("not_FI",[P.P(P.G("ofExn5",[]));P.P(!S.proveIt msg)])))
      | _ -> ())
  | [{D.op = (D.Int 0)}; {D.op = (D.Fun "bool"); D.args = []}] -> 
        raise (S.Contradiction (P.G("not_FI",[P.P(P.G("ofBool0",[]));P.P(!S.proveIt msg)])))
  | [{D.op = (D.Int 1)}; {D.op = (D.Fun "bool"); D.args = []}] -> 
        raise (S.Contradiction (P.G("not_FI",[P.P(P.G("ofBool1",[]));P.P(!S.proveIt msg)])))
  | [x; {D.op = (D.Fun "int"); D.args = []}] -> 
      (match D.nodeVal x with (* convert node to I.exp *)
        D.DExp(nVal) -> 
          raise (S.Contradiction (P.G("not_FI",[P.P(P.G("ofIntAny",[P.E nVal]));P.P(!S.proveIt msg)])))
      | _ -> raise Engine.Failure)
  | [x; {D.op = (D.Fun "top"); D.args = []}] ->
      (match D.nodeVal x with (* convert node to I.exp *)
	D.DExp (nVal) ->
	  raise (S.Contradiction (P.G("not_FI", [P.P(P.G("ofTop", [P.E nVal])); P.P(!S.proveIt msg)])))
      |	_ -> raise Engine.Failure)
  | [{D.op = D.Fun "sel4"; D.args = [m;addr]} as selNode ;t] -> 
      let D.DExp m = D.nodeVal m in
      let D.DExp t = D.nodeVal t in
      
      let (t',g,prfTransOfSel) = findSel(selNode) in
      if (t=t') then
        begin
          let prfTrans  = function glist ->
            P.G("not_FI",[P.P(prfTransOfSel glist); P.P(!S.proveIt msg)]) 
          in
          let sgoal:SG.subgoal = {SG.owner = "TYP"; SG.goal = g; SG.proof_gen = prfTrans; SG.priority =0} in
          SG.add sgoal
        end
      else
        begin
          if (!debug) 
          then ignore (printf "TYP: proving (of sel ...) failed. Return and input types do not match\n");
          
          raise Engine.Failure
        end

  | _ -> 
      if (!debug) then ignore (printf "TYP: processTypingGoal: cannot handle this (of .. ..) predicate ");
      raise Fail
      

(*----------------------------------------------------------------------*)
(* Given a = (of .. ..), see if it is a hypothesis *)

and findAssump (a:D.node)  =
  if (!debug) then ignore (printf "findassump called with %a" D.dn a);
  let rec walkStack x s = 
    match s with
      [] -> E.s (E.bug "TYP: findAssump: no assump found")
    | Snapshot :: rest -> walkStack x rest
    | (Assume msg) :: rest ->
        (match msg with
          {S.body = (S.Literal {D.op = D.BFun "of"; D.args = [v;t]})} ->
              (* found an assump that is (of ... ...) *) 
              (* now check if the variable is the same *)
            if x=v 
            then let D.DExp t = D.nodeVal t in
            (t,[],function [] -> !S.proveIt msg) 
            else walkStack x rest
        | _ -> walkStack x rest)
  in
    (* walk the undoStack and look for an assumption for a *)
  
  match a with
    {D.op = D.Var v} -> walkStack a  !undoStack
  | {D.op = D.Fun v; D.args = []} -> walkStack a  !undoStack
        
  | _ -> raise Fail

(*----------------------------------------------------------------------*)
and findSel (a:D.node) =

  let {D.op=D.Fun "sel4"; D.args=[m;addr]} = a in
  
  let (a,e) = decomposeAddr addr in

  let (t,g,assumpPrf) = findAssump a in (*disallowing anything other than 
                                           * variables for now *)

  match t with
    I.Fun("array",[rw;t;l]) -> 
      
      let sz   = getSizeOnly t in (* get the size of the elements in the 
                                     * array *)
      
      let (e, e',i) = decomposeIndex e sz in (* E' = E * SZ + OFFSET *)
      
      (* now construct the subgoals *)
      let D.DExp addr = D.nodeVal addr in
      let D.DExp a = D.nodeVal a in
      
      let sg0 = I.Atom(I.Eq(addr,I.Plus(a, e))) in (* ADDR = A + E *)
      let sg1 = I.Atom(I.Geq(e,(I.Int 0))) in (* ADDR >= 0 *)
      let sg2 = I.Atom(I.Geq(l,e)) in (* >= L ADDR *)
      
      (* figure out how you are going to construct proofs for (arrIdx ...) 
       * and (fieldof ...) *)
      let modPrf =    (* MOD ADDR SZ = OFFSET *)
        P.G("modCt",[P.E(I.Int sz); P.E e; P.E (I.Int i)])  
      in

      let (fieldPrf,t') = (* OFFSET inside the array element points to something 
                      * meaningful *)
        proveFieldAccess t i;
      in

      (* the real subgoal which is a conjunction of several subgoals *)
      let sg:I.pred list = [sg0;sg1;sg2] in 
      let prfTrans = function sg0Prf::sg1Prf::sg2Prf::gPrf ->
        let arrIdxPrf = P.G("aIdx",[P.E(I.Int sz); P.E(I.Int i); P.P sg1Prf; P.P sg2Prf; P.P modPrf]) in 
        let D.DExp m  = D.nodeVal(m) in
        P.G("ofSel",[P.E m; P.P sg0Prf; P.P fieldPrf; P.P arrIdxPrf; P.P (assumpPrf gPrf)])

      in
      
      if (!debug) 
      then begin
        ignore (printf "ofSel:");
        ignore (printf "subgoals are:\n");
        ignore (printf "  %a\n" I.dp sg0);
        ignore (printf "  %a\n" I.dp sg1);
        ignore (printf "  %a\n" I.dp sg2);
        ignore (printf "subproofs are:\n");
        ignore (printf "mod: %a\n" P.dp modPrf);
        ignore (printf "fieldOf: %a\n" P.dp fieldPrf);
      end;
      if (!debug) 
      then begin
        ignore (printf "returning type: %a" I.de t')
      end;
      (t',sg,prfTrans)


(*----------------------------------------------------------------------*)
(* Prove predicates of the form (saferd4 M ADDR) *)
(* --------------------------------------------- *)

        
let processReadGoal msg args =

  let [m;addr] = args in (* args is M and ADDR *)


  let (a,e) = decomposeAddr addr in (* ADDR = A + E *)
  

  (* a could be a variable or it could be a (sel ...)
   * g is the list of subgoals that need to be proved inorder to use the 
   * assumpPrf proof transformer to get the proof that the addr is an array 
   * If a is a variable, then there will be no subgoals, and the proof is 
   * just a hypothesis. 
   * If a is (sel..) then there will be complicated subgoals *)

  let (t,g,assumpPrf) = 
    (match a with
      {D.op = D.Fun "sel4"} -> findSel a 
    | _ -> findAssump a)
  in

  match t with
    I.Fun("array",[rw;t;l]) ->
      let sz   = getSizeOnly t in (* get the size of the elements in the 
                                   * array *)

      let (e, e',i) = decomposeIndex e sz in (* E' = E * SZ + OFFSET *)

      (* now construct the subgoals *)
      let D.DExp addr = D.nodeVal addr in
      let D.DExp a = D.nodeVal a in

      let sg0 = I.Atom(I.Eq(addr,I.Plus(a, e))) in (* ADDR = A + E *)
      let sg1 = I.Atom(I.Geq(e,(I.Int 0))) in (* ADDR >= 0 *)
      let sg2 = I.Atom(I.Geq(l,e)) in (* >= L ADDR *)
      
      (* figure out how you are going to construct proofs for (arrIdx ...) 
       * and (fieldof ...) *)
      let modPrf =    (* MOD ADDR SZ = OFFSET *)
        P.G("modCt",[P.E(I.Int sz); P.E e; P.E (I.Int i)])  
      in

      let (fieldPrf,_) = (* OFFSET inside the array element points to something 
                      * meaningful *)
        if (!debug) then (ignore (printf "calling fieldPrf: t = %a" I.de t);
                          ignore (printf "                  i = %d" i));
        proveFieldAccess t i;
      in

      (* the real subgoal which is a conjunction of several subgoals *)
      let sg:I.pred list = [sg0;sg1;sg2] @ g in 
      let prfTrans = function sg0Prf::sg1Prf::sg2Prf::gPrf ->
        let arrIdxPrf = P.G("aIdx",[P.E(I.Int sz); P.E(I.Int i); P.P sg1Prf; P.P sg2Prf; P.P modPrf]) in 
        let D.DExp m  = D.nodeVal(m) in
        P.G("not_FI",[P.P(P.G("safeRd",[P.E  m ; P.P sg0Prf; P.P fieldPrf; P.P arrIdxPrf
                                           ; P.P (assumpPrf gPrf)]));
                       P.P(!S.proveIt msg)])


      in
      
      (* now construct the real subgoal *)
      let sgoal:SG.subgoal = {SG.owner = "TYP"; SG.goal = sg; SG.proof_gen = prfTrans ; SG.priority =0} in
      if (!debug) 
      then begin
        ignore (printf "subgoals are:\n");
        ignore (printf "  %a\n" I.dp sg0);
        ignore (printf "  %a\n" I.dp sg1);
        ignore (printf "  %a\n" I.dp sg2);
        ignore (printf "subproofs are:\n");
        ignore (printf "mod: %a\n" P.dp modPrf);
        ignore (printf "fieldOf: %a\n" P.dp fieldPrf);
      end;
      SG.add sgoal

  | I.Fun("ptropt",[t;l]) ->

      let sz  = getSizeOnly t in
      let (e, e',i) = decomposeIndex e sz in (*  E = E' * SZ + OFFSET *)
      (* now construct the subgoals *)

      let D.DExp addr = D.nodeVal addr in
      let D.DExp a = D.nodeVal a in

      let sg0 = I.Atom(I.Eq(addr,I.Plus(a, e))) in
      let sg1 = I.Atom(I.Geq(e,(I.Int 0))) in
      let sg2 = I.Atom(I.Geq(l,e)) in
      let sg3 = I.Atom(I.Neq(a,I.Int 0)) in

      (* figure out how you are going to construct proofs for (arrIdx ...) 
       * and (fieldof ...) *)
      let modPrf = (* proof of (= (mod ...) ..) needed for arrIdx *)
            P.G("modCt",[P.E (I.Int sz); P.E e; P.E (I.Int i)])
      in
      
      let (fieldPrf,_) = (* proof of (fieldOf ..) needed for arrIdx *)
        proveFieldAccess t i
      in

      (* the real subgoal which is a conjunction of several subgoals *)
      let sg = [sg0;sg1;sg2;sg3] in
      let prfTrans = function (sg0Prf::sg1Prf::sg2Prf::sg3Prf::gPrf) ->
        let arrIdxPrf = P.G("aIdx",[P.E(I.Int sz); P.E (I.Int i); P.P sg1Prf; P.P sg2Prf; P.P modPrf]) in 
        let D.DExp m = D.nodeVal(m) in
        P.G("not_FI",[P.P(P.G("safeRd",[P.E m; P.P sg0Prf; P.P fieldPrf; P.P arrIdxPrf; 
                       P.P(P.G("ofPtrOpt",[P.P sg3Prf; P.P (assumpPrf gPrf)]))]));
                      P.P(!S.proveIt msg)])
      in
      
      (* now construct the real subgoal *)
      let sgoal:SG.subgoal = {SG.owner = "TYP"; SG.goal = sg; SG.proof_gen = prfTrans; SG.priority =0} in
      if (!debug) 
      then begin
        ignore (printf "subgoals are:\n");
        ignore (printf "  %a\n" I.dp sg0);
        ignore (printf "  %a\n" I.dp sg1);
        ignore (printf "  %a\n" I.dp sg2);
        ignore (printf "subproofs are:\n");
        ignore (printf "mod: %a\n" P.dp modPrf);
        ignore (printf "fieldOf: %a\n" P.dp fieldPrf);
      end;
      SG.add sgoal
      
  | _ -> E.s (E.unimp "TYP: addr != array + something or ptropt + something")
                                

(*----------------------------------------------------------------------*)
let processWriteGoal msg args =

  let [m;addr;v] = args in (* args is M, ADDR and V *)

  let (a,e) = decomposeAddr addr in (* ADDR = A + E *)


  let (t,g,assumpPrf) = 
    (match a with
      {D.op = D.Fun "sel4"} -> findSel a 
    | _ -> findAssump a)
  in

  match t with
    I.Fun("array",[rw;t;l]) ->
      
      let sz   = getSizeOnly t in (* get the size of the elements in the 
                                     * array *)

      let (e, e',i) = decomposeIndex e sz in (* E' = E * SZ + OFFSET *)

      (* now construct the subgoals *)
      let D.DExp addr = D.nodeVal addr in
      let D.DExp a = D.nodeVal a in
      let D.DExp v = D.nodeVal v in

      let sg0 = I.Atom(I.Eq(addr,I.Plus(a, e))) in (* ADDR = A + E *)
      let sg1 = I.Atom(I.Geq(e,(I.Int 0))) in (* ADDR >= 0 *)
      let sg2 = I.Atom(I.Geq(l,e)) in (* >= L ADDR *)

      (* figure out how you are going to construct proofs for (arrIdx ...) 
       * and (fieldof ...) *)
      let modPrf =    (* MOD ADDR SZ = OFFSET *)
        P.G("modCt",[P.E(I.Int sz); P.E e; P.E (I.Int i)])  
      in

      let (fieldPrf,t') = (* OFFSET inside the array element points to something 
                      * meaningful *)
        proveFieldAccess t i;
      in
      let sg3 = I.Atom(I.BFun("of",[v;t'])) in


      (* the real subgoal which is a conjunction of several subgoals *)
      let sg:I.pred list = [sg0;sg1;sg2;sg3] @ g in 
      let prfTrans = function (sg0Prf::sg1Prf::sg2Prf::sg3Prf::gPrf) ->
        let arrIdxPrf = P.G("aIdx",[P.E(I.Int sz); P.E(I.Int i); P.P sg1Prf; P.P sg2Prf; P.P modPrf]) in 
        let D.DExp m  = D.nodeVal(m) in
        P.G("not_FI",[P.P(P.G("safeWr",[P.E  m ; P.P sg3Prf; P.P sg0Prf; P.P fieldPrf; 
                                         P.P arrIdxPrf; P.P (assumpPrf gPrf)]));
                       P.P(!S.proveIt msg)])




      in
      
      (* now construct the real subgoal *)
      let sgoal:SG.subgoal = {SG.owner = "TYP"; SG.goal = sg; SG.proof_gen = prfTrans ; SG.priority =0} in
      if (!debug) 
      then begin
        ignore (printf "subgoals are:\n");
        ignore (printf "  %a\n" I.dp sg0);
        ignore (printf "  %a\n" I.dp sg1);
        ignore (printf "  %a\n" I.dp sg2);
        ignore (printf "subproofs are:\n");
        ignore (printf "mod: %a\n" P.dp modPrf);
        ignore (printf "fieldOf: %a\n" P.dp fieldPrf);
      end;
      SG.add sgoal

  | I.Fun("ptropt",[t;l]) -> 

      let sz  = getSizeOnly t in
      let (e, e',i) = decomposeIndex e sz in (*  E = E' * SZ + OFFSET *)
      (* now construct the subgoals *)

      let D.DExp addr = D.nodeVal addr in
      let D.DExp a = D.nodeVal a in
      let D.DExp v = D.nodeVal v in

      let sg0 = I.Atom(I.Eq(addr,I.Plus(a, e))) in
      let sg1 = I.Atom(I.Geq(e,(I.Int 0))) in
      let sg2 = I.Atom(I.Geq(l,e)) in
      
      let sg3 = I.Atom(I.Neq(a,I.Int 0)) in
      (* figure out how you are going to construct proofs for (arrIdx ...) 
       * and (fieldof ...) *)
      let modPrf = (* proof of (= (mod ...) ..) needed for arrIdx *)
            P.G("modCt",[P.E (I.Int sz); P.E e; P.E (I.Int i)])
      in
      
      let (fieldPrf,t') = (* proof of (fieldOf ..) needed for arrIdx *)
        proveFieldAccess t i
      in
      let sg4 = I.Atom(I.BFun("of",[v;t'])) in


      (* the real subgoal which is a conjunction of several subgoals *)
      let sg = [sg0;sg1;sg2;sg3;sg4] @ g in
      let prfTrans = function sg0Prf::sg1Prf::sg2Prf::sg3Prf::sg4Prf::gPrf  ->
        let arrIdxPrf = P.G("aIdx",[P.E(I.Int sz); P.E (I.Int i); P.P sg1Prf; P.P sg2Prf; P.P modPrf]) in 
        let D.DExp m = D.nodeVal(m) in
        P.G("not_FI",[P.P(P.G("safeWr",[P.E m; P.P sg4Prf; P.P sg0Prf; P.P fieldPrf; P.P arrIdxPrf; 
                       P.P(P.G("ofPtrOpt",[P.P sg3Prf; P.P (assumpPrf gPrf)]))]));
                      P.P(!S.proveIt msg)])
      in
      
      (* now construct the real subgoal *)
      let sgoal:SG.subgoal = {SG.owner = "TYP"; SG.goal = sg; SG.proof_gen = prfTrans; SG.priority =0} in
      if (!debug) 
      then begin
        ignore (printf "subgoals are:\n");
        ignore (printf "  %a\n" I.dp sg0);
        ignore (printf "  %a\n" I.dp sg1);
        ignore (printf "  %a\n" I.dp sg2);
        ignore (printf "subproofs are:\n");
        ignore (printf "mod: %a\n" P.dp modPrf);
        ignore (printf "fieldOf: %a\n" P.dp fieldPrf);
      end;
      SG.add sgoal
      
  | _ -> E.s (E.unimp "TYP: addr != array + something or ptropt + something")
                                

(*----------------------------------------------------------------------*)
(* Receives a goal message and decides what to do about it, depending on 
 * whether it is (of ... ...), (saferd4 ... ...), (safewr4 ... ...) *)
let processGoal msg = 
    match msg.S.body with
      S.Equality (_,_) -> ()
    | S.Literal x -> 
        (match x.D.op with
          D.BFun "of" -> 
            if ((List.length x.D.args) <> 2) 
            then E.s (E.bug "TYP: OF has arity <> 2");

            (try
              processTypingGoal msg x.D.args
            with _ -> ())

        | D.BFun "saferd4" -> 
            (try 
              processReadGoal msg x.D.args
            with _ -> ())
        | D.BFun "safewr4" -> 
            (try 
              processWriteGoal msg x.D.args
            with _ -> ())
        | _ -> ())

(*----------------------------------------------------------------------*)    
let assume acc msg = 
  if (!debug) then (ignore ( printf "TYP: Assume called: "); ignore ( printf "   %a\n" S.dm msg));
  if (msg.S.goal = false)
  then 
    (if (typingAss msg.S.body) (* store all typing assumptions *)

    then 
      begin
        if (!debug) then ignore (printf "TYP: Assumption is about typing\n");
        push (Assume msg)
      end)
  else processGoal msg;
  acc

(*----------------------------------------------------------------------*)
(* initializer. called from main -- that is when id is provided *)
let initialize id = 
  typId := id;
  { S.snapshot = snapshot;
    S.undo = undo;
    S.assume = assume;
    S.proveIt = proveIt;
    S.debug = debug;
    S.ident  = id;
    S.reinitialize = function () -> ();
  } 
    











