
(*
BLAST is a tool for software model checking.
This file is part of BLAST.

Copyright (c) 2002-2007, The BLAST Team.
All rights reserved. 

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. Neither the name of the authors nor their organizations 
   may be used to endorse or promote products
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

(This is the Modified BSD License, see also
 http://www.opensource.org/licenses/bsd-license.php)

The BLAST Team consists of
        Dirk Beyer (SFU), Thomas A. Henzinger (EPFL),
        Ranjit Jhala (UCSD), and Rupak Majumdar (UCLA).

BLAST web page:
        http://mtc.epfl.ch/blast/

Bug reports:
        Dirk Beyer:      firstname.lastname@sfu.ca or
        Rupak Majumdar:  firstname@cs.ucla.edu or
        Ranjit Jhala:    lastname@cs.ucla.edu
 *)
let foci_bin = "foci.opt"

(*NOTE  the ocaml implementation was abandonned, because of the way threads works in ocaml.
        or me (DZ) not able to get a clean implementation of it in ocaml.
        (the second is more likely)
*)
(*
TODO problably forgetting to close all the file handle somewhere
*)

type foci_struct = {
    bin: string;
    args: string array;
    file_in: string;
    file_out: string;
    ready: Condition.t;
    used: Condition.t;
    reading: Condition.t;
    mutex: Mutex.t;
    pid: int ref;
    is_ready: bool ref;
    is_reading: bool ref
}

let foci = {
    bin       = foci_bin;
    args      = [|foci_bin; "foci_in"; "foci_out"|];
    file_in   = "foci_in";
    file_out  = "foci_out";
    ready     = Condition.create ();
    used      = Condition.create ();
    reading   = Condition.create ();
    mutex     = Mutex.create ();
    pid       = ref 0;
    is_ready  = ref false;
    is_reading= ref false
}
let foci_sym = {
    bin       = foci_bin;
    args      = [|foci_bin; "-s"; "foci_sym_in"; "foci_sym_out"|];
    file_in   = "foci_sym_in";
    file_out  = "foci_sym_out";
    ready     = Condition.create ();
    used      = Condition.create ();
    reading   = Condition.create ();
    mutex     = Mutex.create ();
    pid       = ref 0;
    is_ready  = ref false;
    is_reading= ref false
}

(*TODO make them unique using the pid*)
let mk_named_pipes foci =
  begin
    try
      ignore(Unix.mkfifo foci.file_in 0o600)
    with _ -> ()
  end;
  begin
    try
      ignore(Unix.mkfifo foci.file_out 0o600)
    with _ -> ();
  end

(*thread charged of watching for foci_pid Errors*)
let foci_killer foci foci_err =
  let ic = Unix.in_channel_of_descr foci_err in
    try
      let line = input_line ic in (*blocking call, read -> foci error*)
        ignore(Sys.command ( "kill " ^ string_of_int !(foci.pid)));
        (*attendre que principale ouvre le pipe*)
        prerr_endline "killer_thread: locking mutex";
        Mutex.lock foci.mutex;
        while !(foci.is_reading) = false do
          prerr_endline "waiting for main thread to read from foci";
          Condition.wait foci.reading foci.mutex;
          prerr_endline "waiting for main thread to read from foci: DONE";
        done;
        Mutex.unlock foci.mutex;
        prerr_endline "killer_thread: mutex unlocked";
        Thread.delay 0.1;(*Hack to avoid the main thread getting stuck*)
        ignore(Sys.command ("echo -n > " ^ foci.file_out));
        print_endline line;
		flush stdout;
        close_in_noerr ic;
    with _ -> ()

(*create process such that they are ready as soon as needed*)
let create_foci is_sym =
  let debug_foci_counter = ref 0 in
  let process foci = 
    while true do
      (*create*)
      let ic, oc = Unix.pipe () in
      let child_pid = Unix.create_process
          foci.bin foci.args
          (Unix.openfile "/dev/null" [Unix.O_RDONLY] 0o666)
          (Unix.openfile "/dev/null" [Unix.O_WRONLY] 0o666)
          oc
      in
      debug_foci_counter := !debug_foci_counter + 1;
      let _ = Thread.create (foci_killer foci) ic in
        (***********************************)
        (*sychronised part with side-effect*)
        prerr_endline "create_thread: locking mutex";
        Mutex.lock foci.mutex;
        foci.pid := child_pid;
        foci.is_ready := true;
        prerr_endline ("signaling foci ready: pid = "^ string_of_int child_pid);
        prerr_endline ("signaling foci ready: request = "^ string_of_int !debug_foci_counter);
        Condition.signal foci.ready;
        (*wait for foci to be used*)
        prerr_endline "waiting for foci to be used";
        Condition.wait foci.used foci.mutex;
        prerr_endline "waiting for foci to be used: DONE";
        (*waiting allows the main thread to give queries*)
        Mutex.unlock foci.mutex;
        prerr_endline "create_thread: mutex unlocked";
        (*end of synchronised part*)
        (***********************************)
        Unix.close ic;
        Unix.close oc;
        ignore( Unix.waitpid [] child_pid)
        (*now foci is dead*)
    done
  in
  if is_sym then
    process foci_sym
  else
    process foci

(*clean exit*)
let terminate () =
  Mutex.lock foci.mutex;
  Mutex.lock foci_sym.mutex;
  (*
  ignore (Sys.command ("kill " ^ string_of_int !(foci.pid) ));
  ignore (Sys.command ("kill " ^ string_of_int !(foci_sym.pid) ));
  *)
  ignore (Sys.command ("killall "^ foci_bin ));
  ignore (Sys.command "rm foci_in foci_out foci_sym_in foci_sym_out ");
  print_endline "fociServer process finished";
  exit 0

(*main task: forwarding the queries*)
let read_from_stdin () =
  let process_query foci query =
    begin
      (***********************************)
      (*synchronised part to check if foci is ready*)
      prerr_endline "main_thread (getting foci): locking mutex";
      Mutex.lock foci.mutex;
      while !(foci.is_ready) = false do
        prerr_endline "waiting for foci to be ready";
        Condition.wait foci.ready foci.mutex;
        prerr_endline "waiting for foci to be ready: DONE";
      done;
      foci.is_ready := false;
      Mutex.unlock foci.mutex;
      prerr_endline "main_thread (getting foci): mutex unlocked";
      (*end of synchronised part*)
      (***********************************)
      try
        let oc = open_out foci.file_in in
          output_string oc query;
          output_char oc '\n';
          flush oc;
          close_out oc;
          prerr_endline "main_thread (reading): locking mutex";
          Mutex.lock foci.mutex;
          foci.is_reading := true;
          prerr_endline "signaling beginning to read foci output";
          Condition.signal foci.reading;
          Mutex.unlock foci.mutex;
          prerr_endline "main_thread (reading): mutex unlocked";
          (* this may lead to a deadlock when there are errors:
             the reading condition is used to 'unstuck' the thread when there are errors.
             but, as the open_in call is blocking, it cannot be done mutual exclusion.
             this means the other thread may process between the unlock and the open_in.
          *)
          let ic = open_in foci.file_out in
            try
              while true do
                print_endline (input_line ic)
              done
            with _ ->
              flush stdout;
              close_in_noerr ic;
              (***********************************)
              (*synchronised part to signal foci was used*)
              prerr_endline "main_thread (foci used): locking mutex";
              Mutex.lock foci.mutex;
              prerr_endline "signaling foci was used";
              Condition.signal foci.used;
              Mutex.unlock foci.mutex;
              prerr_endline "main_thread (foci used): mutex unlocked";
              (***********************************)
      with _ ->
        Mutex.unlock foci.mutex;
        prerr_endline "error: during the forwarding of normal query"
      end
  in
  while true do
    let line = input_line stdin in
      match line with
      | "" | "exit" ->
        terminate ()
      | _ ->
        begin
          let query = String.sub line 1 (String.length line -1) in
            match String.get line 0 with
            | 's' -> (*symetric interpolation*)
              process_query foci_sym query
            | 'n' -> (*normal interpolation (inductive)*)
              process_query foci query
            |  _  ->
              begin
                prerr_endline "unknown query type";
                terminate()
              end
        end
  done

let set_signalHandler =
  let _ = Sys.signal Sys.sigabrt (Sys.Signal_handle (fun _ -> terminate ())) in
  let _ = Sys.signal Sys.sighup  (Sys.Signal_handle (fun _ -> terminate ())) in
  let _ = Sys.signal Sys.sigint  (Sys.Signal_handle (fun _ -> terminate ())) in
  let _ = Sys.signal Sys.sigquit (Sys.Signal_handle (fun _ -> terminate ())) in
  let _ = Sys.signal Sys.sigsegv (Sys.Signal_handle (fun _ -> terminate ())) in
  let _ = Sys.signal Sys.sigterm (Sys.Signal_handle (fun _ -> terminate ())) in
  ()


(*launch everything*)
let main () =
  mk_named_pipes foci;
  mk_named_pipes foci_sym;
  let _ = Thread.create create_foci false in
  let _ = Thread.create create_foci true in
    read_from_stdin ();;

main () ;;
